#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from enum import Enum
from typing import List, Dict
from uuid import uuid4

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_slice import ABCSlice
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin


class PoaStates(Enum):
    """
    POA states
    """
    Nascent = 1
    SentToAuthority = 2
    Performing = 3
    AwaitingCompletion = 4
    Success = 5
    Failed = 6
    Unknown = 7
    None_ = 8

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def list_values(cls) -> List[int]:
        return list(map(lambda c: c.value, cls))

    @staticmethod
    def translate_list(states: List[str]) -> List[int] or None:
        result = PoaStates.list_values()

        if states is None or len(states) == 0:
            return result

        incoming_states = list(map(lambda x: x.lower(), states))

        for s in PoaStates:
            if s.name.lower() not in incoming_states:
                result.remove(s.value)

        return result


class Poa:
    """
    Represents POA issued to a sliver
    """
    def __init__(self, *, poa_id: str = uuid4().__str__(), operation: str, reservation: ABCReservationMixin = None,
                 sliver_id: ID = None, vcpu_cpu_map: List[Dict[str, str]] = None, node_set: List[str] = None,
                 keys: List[str] = None, bdf: List[str] = None):
        self.poa_id = poa_id
        self.operation = operation
        self.state = PoaStates.Nascent
        self.reservation = reservation
        self.sliver_id = sliver_id
        self.vcpu_cpu_map = vcpu_cpu_map
        self.bdf = bdf
        self.node_set = node_set
        self.keys = keys
        # Sequence number for outgoing poa messages. Increases with every new message.
        self.sequence_poa_out = 0
        # Sequence number for incoming poa messages.
        self.sequence_poa_in = 0
        self.service_pending = PoaStates.None_

        self.slice_object = None
        self.slice_id = None
        self.actor = None
        self.logger = None
        self.dirty = False
        self.info = {}
        self.error_code = 0
        self.message = ""

        if reservation is not None:
            self.sliver_id = reservation.get_reservation_id()
            if reservation.get_slice() is not None:
                self.slice_id = reservation.get_slice().get_slice_id()
                self.slice_object = reservation.get_slice()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['slice_object']
        del state['reservation']
        del state['actor']
        del state['logger']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.reservation = None
        self.slice_object = None
        self.actor = None
        self.logger = None

    def restore(self, *, actor: ABCActorMixin, reservation: ABCReservationMixin):
        """
        Update the reference objects not saved in the database
        @param actor actor
        @param reservation reservation
        """
        self.reservation = reservation
        self.actor = actor
        if actor:
            self.logger = self.actor.get_logger()

        # Update slice/sliver info if available
        if reservation is not None:
            self.sliver_id = reservation.get_reservation_id()
            if reservation.get_slice() is not None:
                self.slice_id = reservation.get_slice().get_slice_id()
                self.slice_object = self.reservation.get_slice()

    def get_error_code(self) -> int:
        return self.error_code

    def get_message(self) -> str:
        return self.message

    def get_poa_id(self) -> str:
        return self.poa_id

    def get_operation(self) -> str:
        return self.operation

    def get_sliver_id(self) -> ID:
        return self.sliver_id

    def get_slice_id(self) -> ID:
        return self.slice_id

    def get_actor(self) -> ABCActorMixin:
        return self.actor

    def get_state(self) -> PoaStates:
        return self.state

    def get_poa_sequence_in(self) -> int:
        return self.sequence_poa_in

    def get_poa_sequence_out(self) -> int:
        return self.sequence_poa_out

    def get_slice(self) -> ABCSlice:
        return self.slice_object

    def get_reservation(self) -> ABCReservationMixin:
        return self.reservation

    def get_info(self) -> dict:
        return self.info

    def is_dirty(self) -> bool:
        return self.dirty

    def clear_dirty(self):
        self.dirty = False

    def set_dirty(self):
        self.dirty = True

    def transition(self, *, prefix: str, state: PoaStates):
        """
        Transition states for the POA
        @param prefix prefix message to be logged with the transition
        @param state state for the POA
        """
        if self.state == PoaStates.Failed and self.logger is not None:
            self.logger.debug("POA is marked as failed!")

        if self.logger is not None:
            self.logger.debug(f"POA #{self.poa_id} {prefix} transition: {self.get_state()} -> {state.name}")

        change = self.state != state

        if change:
            self.state = state
            self.set_dirty()

    def send_poa_to_authority(self):
        """
        Send POA to the Authority which owns the sliver
        """
        from fabric_cf.actor.core.kernel.reservation_client import ReservationClient
        if not isinstance(self.reservation, ReservationClient):
            raise Exception("POA can be triggered only from Orchestrator")

        self.sequence_poa_out += 1
        from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
        RPCManagerSingleton.get().poa(proxy=self.reservation.get_authority(), poa=self,
                                      callback=self.reservation.get_client_callback_proxy(),
                                      caller=self.slice_object.get_owner())

        # Transition to SentToAuthority
        self.transition(prefix=f"Issued POA to {self.reservation.get_authority()}", state=PoaStates.SentToAuthority)

    def send_poa_info_to_orchestrator(self):
        """
        Send POA Response back to orchestrator
        """
        from fabric_cf.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
        self.sequence_poa_out += 1
        RPCManagerSingleton.get().poa_info(reservation=self.reservation, poa=self)

    def fail(self, *, message: str, notify: bool = False):
        """
        Move POA to failed state and notify orchestrator
        """
        self.transition(prefix=message, state=PoaStates.Failed)
        if notify:
            self.send_poa_info_to_orchestrator()

    def process_poa_authority(self):
        """
        Process POA for a sliver at Authority
        """
        from fabric_cf.actor.core.kernel.authority_reservation import AuthorityReservation
        if not isinstance(self.reservation, AuthorityReservation):
            raise Exception("POA can be processed only at Authority")

        # Transition to Performing State
        self.transition(prefix=f"Performing POA", state=PoaStates.Performing)

    def accept_poa_info(self, *, incoming: Poa):
        """
        Accept POA response from Authority at Orchestrator
        @param incoming incoming POA received from Authority
        """
        # Transition to Success state
        if incoming.error_code == 0:
            self.transition(prefix="done", state=PoaStates.Success)
            from fabric_cf.actor.core.common.event_logger import EventLoggerSingleton
            from fabric_cf.actor.core.proxies.kafka.translate import Translate
            EventLoggerSingleton.get().log_sliver_event(
                slice_object=Translate.translate_slice_to_avro(slice_obj=self.reservation.get_slice()),
                sliver=self.get_reservation().get_resources().get_sliver(), verb=f"poa-{self.operation}",
                keys=self.keys)

            # Copy any information returned
            if incoming.get_info() is not None:
                self.info = incoming.get_info().copy()
        else:
            # Copy the error message
            self.error_code = incoming.get_error_code()
            self.message = incoming.get_message()
            # Fail the POA
            self.fail(message=f"POA failed: {incoming.get_message()}")

    def accept_authority_poa_info(self, *, poa_info: dict):
        """
        Accept the response from Authority Handler execution and send the response to orchestrator
        @param poa_info poa information returned by the handler
        @param notice success/error message returned by the handler
        """
        # Grab the info dictionary
        if poa_info is not None and poa_info.get(Constants.PROPERTY_INFO) is not None:
            self.info = poa_info.get(Constants.PROPERTY_INFO)

        # Get the error code
        self.error_code = poa_info.get(Constants.PROPERTY_CODE)

        # Get any success/error messages
        if poa_info.get(Constants.PROPERTY_MESSAGE) is not None:
            self.message = poa_info.get(Constants.PROPERTY_MESSAGE)

        # send poa info back to the orchestrator
        self.send_poa_info_to_orchestrator()

    def is_failed(self):
        return self.state == PoaStates.Failed

    def is_issued(self):
        return self.state == PoaStates.SentToAuthority

    def is_performing(self):
        return self.state == PoaStates.Performing

    def is_awaiting_response(self):
        return self.state == PoaStates.AwaitingCompletion

    def service_poa(self):
        # Transition to AwaitingCompletion state after handler has been invoked
        self.transition(prefix="Triggered POA to the Handler", state=PoaStates.AwaitingCompletion)

    def clone(self):
        """
        Clone the object
        """
        result = Poa(operation=self.operation, poa_id=self.poa_id, reservation=self.reservation,
                     sliver_id=self.sliver_id)

        if self.vcpu_cpu_map is not None:
            result.vcpu_cpu_map = self.vcpu_cpu_map.copy()

        if self.node_set is not None:
            result.node_set = self.node_set.copy()

        if self.info is not None:
            result.info = self.info.copy()

        if self.keys is not None:
            result.keys = self.keys.copy()

        if self.bdf is not None:
            result.bdf = self.bdf.copy()

        result.error_code = self.error_code

        if self.reservation is not None:
            result.sliver_id = self.reservation.get_reservation_id()
            if self.reservation.get_slice() is not None:
                result.slice_id = self.reservation.get_slice().get_slice_id()
                result.slice_object = self.reservation.get_slice()

        result.state = self.state
        result.sequence_poa_out = self.sequence_poa_out
        result.sequence_poa_in = self.sequence_poa_in
        result.actor = self.actor
        result.logger = self.logger

    def to_dict(self) -> dict:
        """
        Translate POA to a dict object; used to send the information to the handler
        """
        result = {'operation': self.operation, 'poa_id': self.poa_id}
        if self.vcpu_cpu_map is not None:
            result['vcpu_cpu_map'] = self.vcpu_cpu_map
        if self.node_set is not None:
            result['node_set'] = self.node_set
        if self.keys is not None:
            result['keys'] = self.keys
        if self.bdf is not None:
            result['bdf'] = self.bdf
        return result

    def __str__(self):
        return f"POA ID# {self.poa_id} operation# {self.operation} in state: {self.state}"


class PoaFactory:
    @staticmethod
    def create(*, poa_id: str, operation: str, sliver_id: ID, vcpu_cpu_map: List[Dict[str, str]] = None,
               node_set: List[str] = None, keys: List[Dict[str, str]] = None,
               bdf: List[str] = None) -> Poa:
        """
        Create POA
        :param poa_id:
        :param operation:
        :param sliver_id:
        :param vcpu_cpu_map:
        :param node_set:
        :param keys:
        :param bdf:
        :return:
        """
        result = Poa(poa_id=poa_id, operation=operation, vcpu_cpu_map=vcpu_cpu_map, node_set=node_set,
                     sliver_id=sliver_id, keys=keys, bdf=bdf)
        return result



from typing import List

from fabric_mb.message_bus.messages.poa_avro import PoaAvro

from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_handler import OrchestratorHandler
from fabric_cf.orchestrator.swagger_server import received_counter, success_counter, failure_counter
from fabric_cf.orchestrator.swagger_server.models import PoaData
from fabric_cf.orchestrator.swagger_server.models.poa import Poa  # noqa: E501
from fabric_cf.orchestrator.swagger_server.models.poa_post import PoaPost  # noqa: E501
from fabric_cf.orchestrator.swagger_server.response.constants import POST_METHOD, POAS_POST_SLIVER_ID_PATH, \
    POAS_GET_PATH, POAS_GET_POA_ID_PATH
from fabric_cf.orchestrator.swagger_server.response.utils import get_token, cors_success_response, cors_error_response


def poas_create_sliver_id_post(body: PoaPost, sliver_id: str):  # noqa: E501
    """Perform an operational action on a sliver.

    Request to perform an operation action on a sliver. Supported actions include - reboot a VM sliver, get cpu info, get numa info, pin vCPUs, pin memory to a numa node etc.    # noqa: E501

    :param body: Perform Operation Action
    :type body: dict | bytes
    :param sliver_id: Sliver identified by universally unique identifier
    :type sliver_id: str

    :rtype: Poa
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, POAS_POST_SLIVER_ID_PATH).inc()
    try:
        logger.info(f"KOMAL -- incoming: {body.data}")
        token = get_token()
        poa_avro = PoaAvro(operation=body.operation, rid=sliver_id)
        if body.data is not None:
            poa_avro.node_set = body.data.node_set
            poa_avro.vcpu_cpu_map = body.data.vcpu_cpu_map
            poa_avro.keys = body.data.keys
            poa_avro.bdf = body.data.bdf
        poa_id, slice_id = handler.poa(sliver_id=sliver_id, token=token, poa=poa_avro)
        poa_data = PoaData(poa_id=poa_id, operation=body.operation,
                           sliver_id=sliver_id, slice_id=slice_id)
        response = Poa()
        response.data = [poa_data]
        response.size = len(response.data)
        response.type = body.operation
        success_counter.labels(POST_METHOD, POAS_POST_SLIVER_ID_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, POAS_POST_SLIVER_ID_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, POAS_POST_SLIVER_ID_PATH).inc()
        return cors_error_response(error=e)


def poas_get(sliver_id: str =None, states: List[str] = None, limit: int = None, offset: int = None):  # noqa: E501
    """Request get the status of the POAs.

    Request get the status of the POAs    # noqa: E501

    :param sliver_id: Search for POAs for a sliver
    :type sliver_id: str
    :param states: Search for POAs in the specified states
    :type states: List[str]
    :param limit: maximum number of results to return per page (1 or more)
    :type limit: int
    :param offset: number of items to skip before starting to collect the result set
    :type offset: int

    :rtype: Poa
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, POAS_GET_PATH).inc()
    try:
        token = get_token()
        poa_list = handler.get_poas(sliver_id=sliver_id, token=token, limit=limit, offset=offset, states=states)
        response = Poa()
        response.data = []
        for p in poa_list:
            poa = PoaData().from_dict(p)
            response.data.append(poa)
        response.size = len(response.data)
        response.type = "poas"
        success_counter.labels(POST_METHOD, POAS_GET_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, POAS_GET_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, POAS_GET_PATH).inc()
        return cors_error_response(error=e)


def poas_poa_id_get(poa_id):  # noqa: E501
    """Perform an operational action on a sliver.

    Request get the status of the POA identified by poa_id.    # noqa: E501

    :param poa_id: Poa Id for the POA triggered
    :type poa_id: str

    :rtype: Poa
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(POST_METHOD, POAS_GET_POA_ID_PATH).inc()
    try:
        token = get_token()
        poa_list = handler.get_poas(token=token, poa_id=poa_id)
        response = Poa()
        response.data = []
        for p in poa_list:
            poa = PoaData().from_dict(p)
            response.data.append(poa)
        response.size = len(response.data)
        response.type = "poas"
        success_counter.labels(POST_METHOD, POAS_GET_POA_ID_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, POAS_GET_POA_ID_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(POST_METHOD, POAS_GET_POA_ID_PATH).inc()
        return cors_error_response(error=e)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, List, Dict

from fabric_cf.actor.boot.configuration_exception import ConfigurationException
from fabric_cf.actor.boot.inventory.aggregate_resource_model_creator import AggregateResourceModelCreator
from fabric_cf.actor.core.apis.abc_authority import ABCAuthority
from fabric_cf.actor.core.apis.abc_policy import ABCPolicy
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.resource_config import ResourceConfig, ResourceConfigBuilder
from fabric_cf.actor.core.container.remote_actor_cache import RemoteActorCacheSingleton, RemoteActorCache
from fabric_cf.actor.core.core.authority import Authority
from fabric_cf.actor.core.core.broker import Broker
from fabric_cf.actor.core.core.controller import Controller
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.manage.management_utils import ManagementUtils
from fabric_cf.actor.core.manage.messages.client_mng import ClientMng
from fabric_cf.actor.core.plugins.base_plugin import BasePlugin
from fabric_cf.actor.core.plugins.handlers.ansible_handler_processor import AnsibleHandlerProcessor
from fabric_cf.actor.core.plugins.handlers.handler_processor import HandlerProcessor
from fabric_cf.actor.core.plugins.db.server_actor_database import ServerActorDatabase
from fabric_cf.actor.core.plugins.substrate.authority_substrate import AuthoritySubstrate
from fabric_cf.actor.core.plugins.substrate.substrate_mixin import SubstrateMixin
from fabric_cf.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric_cf.actor.core.policy.authority_calendar_policy import AuthorityCalendarPolicy
from fabric_cf.actor.core.policy.broker_simpler_units_policy import BrokerSimplerUnitsPolicy
from fabric_cf.actor.core.policy.controller_ticket_review_policy import ControllerTicketReviewPolicy
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.reflection_utils import ReflectionUtils
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.core.core.actor import ActorMixin
from fabric_cf.actor.security.auth_token import AuthToken
from fabric_cf.actor.core.apis.abc_actor_mixin import ActorType

if TYPE_CHECKING:
    from fabric_cf.actor.boot.configuration import Configuration, ActorConfig, PolicyConfig, Peer
    from fabric_cf.actor.core.apis.abc_actor_mixin import ABCActorMixin
    from fabric_cf.actor.core.apis.abc_mgmt_actor import ABCMgmtActor


class ExportAdvertisement:
    """
    Encapsulates information needed for the delegations to be exported which are later claimed by broker
    """
    def __init__(self, *, exporter: ABCMgmtActor, client: ClientMng, delegation: str, topic: str):
        self.exporter = exporter
        self.client = client
        self.delegation = delegation
        self.client_topic = topic
        self.exported = None


class ConfigurationProcessor:
    """
    Processing the Configuration loaded and brings up the required actors, peers and various threads
    """
    def __init__(self, *, config: Configuration):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.config = config
        self.actor = None
        self.to_export = []
        self.to_advertise = []
        self.resources = {}
        self.controls = None
        self.aggregate_delegation_models = None

    def process(self):
        """
        Instantiates the configuration
        @raises ConfigurationException in case of error
        """
        try:
            self.create_actor()
            self.initialize_actor()
            self.register_actor()
            self.create_default_slice()
            self.populate_inventory_neo4j()
            self.recover_actor()
            self.enable_ticking()
            self.process_topology()
            self.process_advertise()
            self.logger.info("Processing exports completed")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            raise ConfigurationException(f"Unexpected error while processing configuration {e}")
        self.logger.info("Finished instantiating actors.")
        print("End process")

    def create_actor(self):
        """
        Instantiates all actors
        @raises ConfigurationException in case of error
        """
        try:
            if self.config.get_actor_config() is not None:
                self.logger.debug(f"Creating Actor: name={self.config.get_actor_config().get_name()}")
                self.actor = self.do_common(actor_config=self.config.get_actor_config())
                self.do_specific(actor=self.actor, config=self.config.get_actor_config())
        except Exception as e:
            raise ConfigurationException(f"Unexpected error while creating actor {e}")

    def do_common(self, *, actor_config: ActorConfig):
        """
        Perform the common operations for actor creation
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        actor = self.make_actor_instance(actor_config=actor_config)
        actor.set_plugin(plugin=self.make_plugin_instance(actor=actor, actor_config=actor_config))
        actor.set_policy(policy=self.make_actor_policy(actor=actor, config=actor_config))
        return actor

    @staticmethod
    def make_actor_instance(*, actor_config: ActorConfig) -> ABCActorMixin:
        """
        Creates Actor instance
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        actor_type = ActorType.get_actor_type_from_string(actor_type=actor_config.get_type())
        actor = None
        if actor_type == ActorType.Orchestrator:
            actor = Controller()
        elif actor_type == ActorType.Broker:
            actor = Broker()
        elif actor_type == ActorType.Authority:
            actor = Authority()
        else:
            raise ConfigurationException(f"Unsupported actor type: {actor_type}")

        actor_guid = ID()
        if actor_config.get_guid() is not None:
            actor_guid = ID(uid=actor_config.get_guid())
        auth_token = AuthToken(name=actor_config.get_name(), guid=actor_guid)
        actor.set_identity(token=auth_token)
        if actor_config.get_description() is not None:
            actor.set_description(description=actor_config.get_description())

        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        actor.set_actor_clock(clock=GlobalsSingleton.get().get_container().get_actor_clock())

        return actor

    def make_plugin_instance(self, *, actor: ABCActorMixin, actor_config: ActorConfig):
        """
        Creates Plugin instance for the Actor
        @param actor actor
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        plugin = None
        if actor.get_plugin() is None:
            if actor.get_type() == ActorType.Authority:
                plugin = AuthoritySubstrate(actor=actor, db=None, handler_processor=AnsibleHandlerProcessor())

            elif actor.get_type() == ActorType.Orchestrator:
                plugin = SubstrateMixin(actor=actor, db=None, handler_processor=HandlerProcessor())

            elif actor.get_type() == ActorType.Broker:
                plugin = BasePlugin(actor=actor, db=None, handler_processor=HandlerProcessor())

        if plugin is None:
            raise ConfigurationException(f"Cannot instantiate plugin for actor: {actor_config.get_name()}")

        if plugin.get_database() is None:
            db = None
            user = self.config.get_global_config().get_database()[Constants.PROPERTY_CONF_DB_USER]
            password = self.config.get_global_config().get_database()[Constants.PROPERTY_CONF_DB_PASSWORD]
            db_host = self.config.get_global_config().get_database()[Constants.PROPERTY_CONF_DB_HOST]
            db_name = self.config.get_global_config().get_database()[Constants.PROPERTY_CONF_DB_NAME]
            if isinstance(plugin, SubstrateMixin):
                db = SubstrateActorDatabase(user=user, password=password, database=db_name, db_host=db_host,
                                            logger=self.logger)
            else:
                db = ServerActorDatabase(user=user, password=password, database=db_name, db_host=db_host,
                                         logger=self.logger)

            plugin.set_database(db=db)
        return plugin

    def make_actor_policy(self, *, actor: ABCActorMixin, config: ActorConfig):
        """
        Creates Actor Policy instance
        @param actor actor
        @param config actor config
        @raises ConfigurationException in case of error
        """
        policy = None
        if actor.get_type() == ActorType.Authority:
            policy = self.make_site_policy(config=config)
        elif actor.get_type() == ActorType.Broker:
            policy = self.make_broker_policy(config=config)
        elif actor.get_type() == ActorType.Orchestrator:
            if config.get_policy() is not None:
                policy = self.make_policy(policy=config.get_policy())
            else:
                policy = ControllerTicketReviewPolicy()

        if policy is None:
            raise ConfigurationException(f"Could not instantiate policy for actor: {config.get_name()}")
        return policy

    def make_broker_policy(self, *, config: ActorConfig):
        """
        Creates AM Policy instance and set up controls
        @param config actor config
        @raises ConfigurationException in case of error
        """
        if config.get_policy() is not None:
            policy = self.make_policy(policy=config.get_policy())
            properties = config.get_policy().get_properties()
            policy.set_properties(properties=properties)
        else:
            policy = BrokerSimplerUnitsPolicy()

        for i in config.get_controls():
            try:
                if i.get_module_name() is None or i.get_class_name() is None:
                    raise ConfigurationException("Missing inventory class name")

                inventory = ReflectionUtils.create_instance(module_name=i.get_module_name(),
                                                            class_name=i.get_class_name())
                inventory.set_logger(logger=self.logger)

                if i.get_type() is None:
                    raise ConfigurationException("No type specified for control")

                for t in i.get_type():
                    policy.register_inventory(resource_type=ResourceType(resource_type=t), inventory=inventory)
            except Exception as e:
                self.logger.error(traceback.format_exc())
                raise ConfigurationException("Could not create control {}".format(e))
        return policy

    def make_site_policy(self, *, config: ActorConfig):
        """
        Creates AM Policy instance and set up controls
        @param config actor config
        @raises ConfigurationException in case of error
        """
        policy = None
        if config.get_policy() is not None:
            policy = self.make_policy(policy=config.get_policy())
        else:
            policy = AuthorityCalendarPolicy()
        for c in config.get_controls():
            try:
                if c.get_module_name() is None or c.get_class_name() is None:
                    raise ConfigurationException("Missing control class name")

                control = ReflectionUtils.create_instance(module_name=c.get_module_name(),
                                                          class_name=c.get_class_name())
                control.set_actor(actor=self.actor)

                if c.get_type() is None:
                    raise ConfigurationException("No type specified for control")

                for t in c.get_type():
                    control.add_type(rtype=ResourceType(resource_type=t))
                policy.register_control(control=control)
            except Exception as e:
                self.logger.error(traceback.format_exc())
                raise ConfigurationException("Could not create control {}".format(e))
        return policy

    @staticmethod
    def make_policy(*, policy: PolicyConfig) -> ABCPolicy:
        """
        Creates Policy Instance
        @param policy policy config
        @raises ConfigurationException in case of error
        """
        if policy.get_class_name() is None or policy.get_module_name() is None:
            raise ConfigurationException("Policy is missing class name")

        return ReflectionUtils.create_instance(module_name=policy.get_module_name(),
                                               class_name=policy.get_class_name())

    def do_specific(self, *, actor: ABCActorMixin, config: ActorConfig):
        """
        Do Actor specify initialization
        @param actor actor
        @param config actor config
        @raises ConfigurationException in case of error
        """
        if isinstance(actor, ABCAuthority):
            self.resources = ResourceConfigBuilder.build_resource_config(config=config)

    def initialize_actor(self):
        """
        Initialize actor
        @raises ConfigurationException in case of error
        """
        try:
            self.actor.initialize(config=self.config.get_actor_config())
        except Exception as e:
            raise ConfigurationException(f"Actor failed to initialize: {self.actor.get_name()} {e}")

    def register_actor(self):
        """
        Register actor
        @raises ConfigurationException in case of error
        """
        try:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            GlobalsSingleton.get().get_container().register_actor(actor=self.actor)
        except Exception as e:
            raise ConfigurationException(f"Could not register actor: {self.actor.get_name()} {e}")

    def create_default_slice(self):
        """
        Create default slice
        @raises ConfigurationException in case of error
        """
        if self.actor.get_type() != ActorType.Authority:
            slice_obj = SliceFactory.create(slice_id=ID(), name=self.actor.get_name())
            slice_obj.set_inventory(value=True)
            try:
                self.actor.register_slice(slice_object=slice_obj)
            except Exception as e:
                self.logger.error(traceback.format_exc())
                raise ConfigurationException(f"Could not create default slice for actor: {self.actor.get_name()} {e}")

    def populate_inventory_neo4j(self):
        """
        Load Aggregate Resource model
        @raises ConfigurationException in case of error
        """
        if isinstance(self.actor, ABCAuthority) and isinstance(self.actor.get_plugin(), AuthoritySubstrate):
            creator = AggregateResourceModelCreator(substrate=self.actor.get_plugin(), resources=self.resources,
                                                    neo4j_config=self.config.get_global_config().get_neo4j_config())
            self.aggregate_delegation_models = creator.process_neo4j(actor_name=self.actor.get_name(),
                                                                     substrate_file=self.config.get_actor_config().
                                                                     get_substrate_file())
            self.actor.set_aggregate_resource_model(aggregate_resource_model=creator.get_arm_graph())

    def recover_actor(self):
        """
        Recover an actor on stateful restart
        @raises ConfigurationException in case of error
        """
        try:
            self.actor.recover()
        except Exception as e:
            raise ConfigurationException(f"Recovery failed for actor: {self.actor.get_name()} e: {e}")

    def enable_ticking(self):
        """
        Enable the Actor tick
        """
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        GlobalsSingleton.get().get_container().register(tickable=self.actor)

    def process_topology(self):
        """
        Set up Peer registry with identity and kafka topic information
        @raises ConfigurationException in case of error
        """
        if self.config.get_peers() is None or len(self.config.get_peers()) == 0:
            self.logger.debug("No peers specified")
        else:
            self.create_proxies(peers=self.config.get_peers())

    def process_advertise(self):
        """
        Set up Aggregate Delegation models for the peers
        @raises ConfigurationException in case of error
        """
        if self.aggregate_delegation_models is None and len(self.to_advertise) > 0:
            raise ConfigurationException("No delegations found in Aggregate Resource Model")

        for ei in self.to_advertise:
            self.advertise(info=ei)

    def create_proxies(self, *, peers: List[Peer]):
        """
        Set up proxies
        @param peers List of the Peer config
        @raises ConfigurationException in case of error
        """
        for e in peers:
            self.process_peer(peer=e)

    def vertex_to_registry_cache(self, *, peer: Peer):
        """
        Add peer to Registry Cache
        @param peer peer config
        @raises ConfigurationException in case of error
        """
        self.logger.debug("Adding vertex for {}".format(peer.get_name()))

        if peer.get_name() is None:
            raise ConfigurationException("Actor must specify a name")

        if peer.get_guid() is None:
            raise ConfigurationException("Actor must specify a guid")

        protocol = Constants.PROTOCOL_LOCAL
        kafka_topic = None
        if peer.get_kafka_topic() is not None:
            protocol = Constants.PROTOCOL_KAFKA
            kafka_topic = peer.get_kafka_topic()

        actor_type = peer.get_type().lower()

        entry = {
            RemoteActorCache.actor_name: peer.get_name(),
            RemoteActorCache.actor_guid: ID(uid=peer.get_guid()),
            RemoteActorCache.actor_type: actor_type,
            RemoteActorCache.actor_protocol: protocol
        }
        if kafka_topic is not None:
            entry[RemoteActorCache.actor_location] = kafka_topic

        RemoteActorCacheSingleton.get().add_partial_cache_entry(guid=ID(uid=peer.get_guid()), entry=entry)

    def process_peer(self, *, peer: Peer):
        """
        Process a peer
        @param peer peer
        @raises ConfigurationException in case of error
        """
        peer_guid = ID(uid=peer.get_guid())
        peer_type = ActorType.get_actor_type_from_string(actor_type=peer.get_type())
        actor_guid = self.actor.get_guid()
        actor_type = self.actor.get_type()

        # peers between actors of same type aren't allowed unless the actors are both brokers
        if peer_type == actor_type and peer_type != ActorType.Broker:
            raise ConfigurationException(
                "Invalid peer type: broker can only talk to broker, orchestrator or site authority")

        container = ManagementUtils.connect(caller=self.actor.get_identity())
        mgmt_actor = container.get_actor(guid=actor_guid)
        if mgmt_actor is None and container.get_last_error() is not None:
            self.logger.error(container.get_last_error())

        self.vertex_to_registry_cache(peer=peer)

        try:
            client = RemoteActorCacheSingleton.get().establish_peer(mgmt_actor=mgmt_actor, peer_guid=peer_guid,
                                                                    peer_type=peer_type)
            self.logger.debug(f"Client returned {client}")
            if client is not None:
                self.parse_exports(peer=peer, client=client, mgmt_actor=mgmt_actor)
        except Exception as e:
            raise ConfigurationException(f"Could not process exports from: {peer.get_guid()} to "
                                         f"{self.actor.get_guid()}. e= {e}")

    def parse_exports(self, *, peer: Peer, client: ClientMng, mgmt_actor: ABCMgmtActor):
        """
        Identify all the delegations to be advertised
        @param peer peer config
        @param client client
        @param mgmt_actor management actor
        """
        if peer.get_delegation() is not None:
            info = ExportAdvertisement(exporter=mgmt_actor, client=client,
                                       delegation=peer.get_delegation(),
                                       topic=peer.get_kafka_topic())
            self.to_advertise.append(info)

    def advertise(self, *, info: ExportAdvertisement):
        """
        Advertise a delegation
        @param info advertisement information
        @raises ConfigurationException in case of error
        """
        self.logger.debug(f"Using Server Actor {info.exporter.__class__.__name__} to export resources")

        delegation = self.aggregate_delegation_models.get(info.delegation, None)

        if delegation is None:
            raise ConfigurationException(f"Aggregate Delegation Model not found for delegation: {info.delegation}")

        client = AuthToken(name=info.client.get_name(),
                           guid=ID(uid=info.client.get_guid()))
        info.exported = info.exporter.advertise_resources(delegation=delegation, delegation_name=info.delegation,
                                                          client=client)

        if info.exported is None:
            raise ConfigurationException(f"Could not export resources from actor: {info.exporter.get_name()} to actor:"
                                         f" {info.client.get_name()} Error = {info.exporter.get_last_error()}")



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import traceback
from datetime import datetime, timedelta, timezone
from http.client import NOT_FOUND, BAD_REQUEST, UNAUTHORIZED
from typing import List, Union

from fabric_mb.message_bus.messages.auth_avro import AuthAvro
from fabric_mb.message_bus.messages.poa_avro import PoaAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.slice_avro import SliceAvro
from fim.graph.networkx_property_graph_disjoint import NetworkXGraphImporterDisjoint
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.network_service import NetworkServiceSliver
from fim.user import GraphFormat
from fim.user.topology import ExperimentTopology

from fabric_cf.actor.core.common.event_logger import EventLoggerSingleton
from fabric_cf.actor.core.kernel.poa import PoaStates
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.fim.fim_helper import FimHelper
from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.common.constants import Constants, ErrorCodes
from fabric_cf.actor.core.kernel.slice_state_machine import SliceState
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.security.fabric_token import FabricToken
from fabric_cf.actor.security.pdp_auth import ActionId
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_slice_wrapper import OrchestratorSliceWrapper
from fabric_cf.orchestrator.core.orchestrator_kernel import OrchestratorKernelSingleton
from fabric_cf.orchestrator.core.response_builder import ResponseBuilder


class OrchestratorHandler:
    def __init__(self):
        self.controller_state = OrchestratorKernelSingleton.get()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.globals = GlobalsSingleton.get()
        self.logger = self.globals.get_logger()
        self.jwks_url = self.globals.get_config().get_oauth_config().get(Constants.PROPERTY_CONF_O_AUTH_JWKS_URL, None)
        self.pdp_config = self.globals.get_config().get_global_config().get_pdp_config()
        self.config = self.globals.get_config()
        self.infrastructure_project_id = self.config.get_runtime_config().get(Constants.INFRASTRUCTURE_PROJECT_ID, None)
        self.total_slice_count_seed = self.config.get_runtime_config().get(Constants.TOTAL_SLICE_COUNT_SEED, 0)
        self.local_bqm = self.globals.get_config().get_global_config().get_bqm_config().get(
                    Constants.LOCAL_BQM, False)
        excluded_projects = self.config.get_runtime_config().get(Constants.EXCLUDED_PROJECTS, "")
        self.excluded_projects = [e.strip() for e in excluded_projects.split(",") if e.strip()]

    def get_logger(self):
        """
        Get Logger
        :return: logger
        """
        return self.logger

    def __authorize_request(self, *, id_token: str, action_id: ActionId,
                            resource: BaseSliver or ExperimentTopology = None,
                            lease_end_time: datetime = None) -> FabricToken:
        """
        Authorize request
        :param id_token:
        :param action_id:
        :param resource:
        :param lease_end_time:
        :return:
        """
        from fabric_cf.actor.security.access_checker import AccessChecker
        fabric_token = AccessChecker.check_access(action_id=action_id, token=id_token, logger=self.logger,
                                                  resource=resource, lease_end_time=lease_end_time)

        if fabric_token.subject is None:
            raise OrchestratorException(http_error_code=UNAUTHORIZED, message="Invalid token")
        return fabric_token

    def get_broker(self, *, controller: ABCMgmtControllerMixin) -> ID:
        """
        Get broker
        :param controller:
        :return:
        """
        try:
            if self.controller_state.get_broker() is not None:
                return self.controller_state.get_broker()

            brokers = controller.get_brokers()
            self.logger.debug(f"Brokers: {brokers}")
            self.logger.error(f"Last Error: {controller.get_last_error()}")
            if brokers is not None:
                result = ID(uid=next(iter(brokers), None).get_guid())
                self.controller_state.set_broker(broker=result)
                return result
        except Exception as e:
            self.logger.error(f"Error occurred: {e}", stack_info=True)

    def discover_broker_query_model(self, *, controller: ABCMgmtControllerMixin, token: str = None,
                                    level: int = 10, graph_format: GraphFormat = GraphFormat.GRAPHML,
                                    force_refresh: bool = False, start: datetime = None,
                                    end: datetime = None, includes: str = None, excludes: str = None,
                                    email: str = None) -> str or None:
        """
        Discover all the available resources by querying Broker
        :param controller Management Controller Object
        :param token Fabric Token
        :param level: level of details
        :param graph_format: Graph format
        :param force_refresh: Force fetching a fresh model from Broker
        :param start: start time
        :param end: end time
        :param includes: comma separated lists of sites to include
        :param excludes: comma separated lists of sites to exclude
        :param email: Email of the user on whose behalf the request is initiated
        :return str or None
        """
        broker_query_model = None
        # Always get Fresh copy for advanced resource requests
        if not start and not end and not includes and not excludes:
            saved_bqm = self.controller_state.get_saved_bqm(graph_format=graph_format, level=level)
            if saved_bqm is not None:
                if not force_refresh and not saved_bqm.can_refresh() and not saved_bqm.refresh_in_progress:
                    broker_query_model = saved_bqm.get_bqm()
                else:
                    saved_bqm.start_refresh()

        # Request the model from Broker as a fallback
        if not broker_query_model:
            broker = self.get_broker(controller=controller)
            if broker is None:
                raise OrchestratorException("Unable to determine broker proxy for this controller. "
                                            "Please check Orchestrator container configuration and logs.")

            self.logger.info(f"Sending Query to broker on behalf of {email} Start: {start}, End: {end}, "
                             f"Force: {force_refresh}, Level: {level}")

            model = controller.get_broker_query_model(broker=broker, id_token=token, level=level,
                                                      graph_format=graph_format, start=start, end=end,
                                                      includes=includes, excludes=excludes)
            if model is None or model.get_model() is None or model.get_model() == '':
                raise OrchestratorException(http_error_code=NOT_FOUND, message=f"Resource(s) not found for "
                                                                               f"level: {level} format: {graph_format}!")

            broker_query_model = model.get_model()

            # Do not update cache for advance requests
            if not start and not end and not includes and not excludes:
                self.controller_state.save_bqm(bqm=broker_query_model, graph_format=graph_format, level=level)

        return broker_query_model

    def list_resources(self, *, level: int, force_refresh: bool = False, start: datetime = None,
                       end: datetime, includes: str = None, excludes: str = None, graph_format_str: str = None,
                       token: str = None, authorize: bool = True) -> str:
        """
        List Resources
        :param token Fabric Identity Token
        :param level: level of details (default set to 1)
        :param force_refresh: force fetching bqm from broker and override the cached model
        :param start: start time
        :param end: end time
        :param includes: comma separated lists of sites to include
        :param excludes: comma separated lists of sites to exclude
        :param graph_format_str: Graph format
        :param authorize: Authorize the request; Not authorized for Portal requests
        :raises Raises an exception in case of failure
        :returns Broker Query Model on success
        """
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"list_resources invoked controller:{controller}")

            graph_format = self.__translate_graph_format(graph_format=graph_format_str) if graph_format_str else GraphFormat.GRAPHML

            if authorize:
                fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.query)
                email = fabric_token.email
            else:
                email = None
            broker_query_model = self.discover_broker_query_model(controller=controller, token=token, level=level,
                                                                  force_refresh=force_refresh, start=start,
                                                                  end=end, includes=includes, excludes=excludes,
                                                                  graph_format=graph_format, email=email)
            return broker_query_model

        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing list_resources e: {e}")
            raise e

    def create_slice(self, *, token: str, slice_name: str, slice_graph: str, ssh_key: str,
                     lease_start_time: datetime = None, lease_end_time: datetime = None,
                     lifetime: int = 24) -> List[dict]:
        """
        Create a slice
        :param token Fabric Identity Token
        :param slice_name Slice Name
        :param slice_graph Slice Graph Model
        :param ssh_key: User ssh key
        :param lease_start_time: Lease Start Time (UTC)
        :param lease_end_time: Lease End Time (UTC)
        :param lifetime: Lifetime of the slice in hours
        :raises Raises an exception in case of failure
        :returns List of reservations created for the Slice on success
        """
        start = time.time()
        slice_id = None
        controller = None
        new_slice_object = None
        asm_graph = None
        topology = None
        try:
            from fabric_cf.actor.security.access_checker import AccessChecker
            fabric_token = AccessChecker.validate_and_decode_token(token=token)
            project, tags, project_name = fabric_token.first_project
            allow_long_lived = True if Constants.SLICE_NO_LIMIT_LIFETIME in tags else False
            start_time, end_time = self.__compute_lease_end_time(lease_end_time=lease_end_time, lifetime=lifetime,
                                                                 allow_long_lived=allow_long_lived, project_id=project)

            controller = self.controller_state.get_management_actor()

            # Validate the slice graph
            create_ts = time.time()
            topology = ExperimentTopology(graph_string=slice_graph, importer=NetworkXGraphImporterDisjoint())
            topology.validate()
            self.logger.info(f"TV validate: TIME= {time.time() - create_ts:.0f}")

            create_ts = time.time()
            asm_graph = FimHelper.get_neo4j_asm_graph(slice_graph=topology.serialize())
            self.logger.info(f"ASM load: TIME= {time.time() - create_ts:.0f}")

            # Authorize the slice
            create_ts = time.time()
            self.__authorize_request(id_token=token, action_id=ActionId.create, resource=topology,
                                     lease_end_time=end_time)
            self.logger.info(f"PDP authorize: TIME= {time.time() - create_ts:.0f}")

            # Check if an Active slice exists already with the same name for the user
            create_ts = time.time()
            if tags is not None and isinstance(tags, list):
                tags = ','.join(tags)
            existing_slices = controller.get_slices(slice_name=slice_name, user_id=fabric_token.uuid,
                                                    project=project)
            self.logger.info(f"GET slices: TIME= {time.time() - create_ts:.0f}")

            if existing_slices is not None and len(existing_slices) != 0:
                for es in existing_slices:
                    slice_state = SliceState(es.get_state())
                    if not SliceState.is_dead_or_closing(state=slice_state):
                        raise OrchestratorException(f"Slice {slice_name} already exists")

            broker = self.get_broker(controller=controller)
            if broker is None:
                raise OrchestratorException("Unable to determine broker proxy for this controller. "
                                            "Please check Orchestrator container configuration and logs.")

            slice_obj = SliceAvro()
            slice_obj.set_slice_name(slice_name)
            slice_obj.set_client_slice(True)
            slice_obj.set_description("Description")
            slice_obj.graph_id = asm_graph.get_graph_id()
            slice_obj.set_config_properties(value={Constants.USER_SSH_KEY: ssh_key,
                                                   Constants.PROJECT_ID: project,
                                                   Constants.TAGS: tags,
                                                   Constants.CLAIMS_EMAIL: fabric_token.email,
                                                   Constants.TOKEN_HASH: fabric_token.token_hash})
            slice_obj.set_lease_start(lease_start=start_time)
            slice_obj.set_lease_end(lease_end=end_time)
            auth = AuthAvro()
            auth.name = self.controller_state.get_management_actor().get_name()
            auth.guid = self.controller_state.get_management_actor().get_guid()
            auth.oidc_sub_claim = fabric_token.uuid
            auth.email = fabric_token.email
            auth.token = token
            slice_obj.set_owner(auth)
            slice_obj.set_project_id(project)
            slice_obj.set_project_name(project_name)

            create_ts = time.time()
            self.logger.debug(f"Adding Slice {slice_name}")
            slice_id = controller.add_slice(slice_obj=slice_obj)
            self.logger.info(f"SLC add slices: TIME= {time.time() - create_ts:.0f}")
            if slice_id is None:
                self.logger.error(controller.get_last_error())
                self.logger.error("Slice could not be added to Database")
                raise OrchestratorException("Slice could not be added to Database")
            self.logger.debug(f"Slice {slice_name}/{slice_id} added successfully")

            slice_obj.set_slice_id(slice_id=str(slice_id))
            new_slice_object = OrchestratorSliceWrapper(controller=controller, broker=broker,
                                                        slice_obj=slice_obj, logger=self.logger)

            new_slice_object.lock()

            # Create Slivers from Slice Graph; Compute Reservations from Slivers;
            computed_reservations = new_slice_object.create(slice_graph=asm_graph,
                                                            lease_start_time=lease_start_time,
                                                            lease_end_time=lease_end_time,
                                                            lifetime=lifetime)
            new_slice_object.update_topology(topology=topology)

            # Check if Testbed in Maintenance or Site in Maintenance
            self.check_maintenance_mode(token=fabric_token, reservations=computed_reservations)

            create_ts = time.time()
            #if (lease_start_time and lease_end_time and lifetime) or new_slice_object.ero:
            if lease_start_time and lease_end_time and lifetime:
                # Enqueue slices with ERO also on Advanced Scheduling Thread to determine the links
                # Enqueue future slices on Advanced Scheduling Thread to determine possible start time
                # Determining start time may take time so this is done asynchronously to avoid increasing response time
                # of create slice API
                self.controller_state.get_advance_scheduling_thread().queue_slice(controller_slice=new_slice_object)
            else:
                # Enqueue the slice on the demand thread
                # Demand thread is responsible for demanding the reservations
                # Helps improve the create response time

                # Add Reservations to relational database;
                new_slice_object.add_reservations()
                self.logger.info(f"OC wrapper: TIME= {time.time() - create_ts:.0f}")
                self.controller_state.get_defer_thread().queue_slice(controller_slice=new_slice_object)
                self.logger.info(f"QU queue: TIME= {time.time() - create_ts:.0f}")

            EventLoggerSingleton.get().log_slice_event(slice_object=slice_obj, action=ActionId.create,
                                                       topology=topology)

            controller.increment_metrics(project_id=project, oidc_sub=fabric_token.uuid)
            return ResponseBuilder.get_reservation_summary(res_list=computed_reservations)
        except Exception as e:
            if slice_id is not None and controller is not None and asm_graph is not None:
                FimHelper.delete_graph(graph_id=asm_graph.graph_id)
                controller.remove_slice(slice_id=slice_id)
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing create_slice e: {e}")
            raise e
        finally:
            if topology is not None and topology.graph_model is not None:
                topology.graph_model.delete_graph()
            if new_slice_object is not None:
                new_slice_object.unlock()
            self.logger.info(f"OH : TIME= {time.time() - start:.0f}")

    def get_slivers(self, *, token: str, slice_id: str, sliver_id: str = None, as_self: bool = True) -> List[dict]:
        """
        Get Slivers for a Slice
        :param token Fabric Identity Token
        :param slice_id Slice Id
        :param sliver_id Sliver Id
        :param as_self flag; True - return calling user's slivers otherwise, return all slivers in the project
        :raises Raises an exception in case of failure
        :returns List of reservations created for the Slice on success
        """
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"get_slivers invoked for Controller: {controller}")

            slice_guid = ID(uid=slice_id) if slice_id is not None else None
            rid = ID(uid=sliver_id) if sliver_id is not None else None

            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.query)

            # Filter slices based on user's user_id only when querying as_self
            user_id = fabric_token.uuid
            if not as_self:
                user_id = None

            reservations = controller.get_reservations(slice_id=slice_guid, rid=rid, oidc_claim_sub=user_id)
            if reservations is None:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                    if controller.get_last_error().status.code == ErrorCodes.ErrorNoSuchSlice:
                        raise OrchestratorException(f"Slice# {slice_id} not found",
                                                    http_error_code=NOT_FOUND)
                    elif controller.get_last_error().status.code == ErrorCodes.ErrorNoSuchReservation:
                        raise OrchestratorException(f"Reservation# {rid} not found",
                                                    http_error_code=NOT_FOUND)

                raise OrchestratorException(f"Slice# {slice_id} has no reservations",
                                            http_error_code=NOT_FOUND)

            return ResponseBuilder.get_reservation_summary(res_list=reservations)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing get_slivers e: {e}")
            raise e

    def get_slices(self, *, token: str, states: List[str], name: str, limit: int, offset: int,
                   as_self: bool = True, search: str = None, exact_match: bool = False) -> List[dict]:
        """
        Get User Slices
        :param token Fabric Identity Token
        :param states Slice states
        :param name Slice name
        :param limit Number of slices to return
        :param offset Offset
        :param as_self flag; True - return calling user's slices otherwise, return all slices in the project
        :param search: search term applied
        :param exact_match: Exact Match for Search term
        :raises Raises an exception in case of failure
        :returns List of Slices on success
        """
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"get_slices invoked for Controller: {controller}")

            slice_states = SliceState.translate_list(states=states)

            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.query)

            projects = fabric_token.projects
            project = None
            if len(projects) == 1:
                project, tags, project_name = fabric_token.first_project
            else:
                as_self = True

            # Filter slices based on user's user_id only when querying as_self
            user_id = fabric_token.uuid
            if not as_self:
                user_id = None
            slice_list = controller.get_slices(states=slice_states, user_id=user_id, project=project,
                                               slice_name=name, limit=limit, offset=offset)
            return ResponseBuilder.get_slice_summary(slice_list=slice_list)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing get_slices e: {e}")
            raise e

    def modify_slice(self, *, token: str, slice_id: str, slice_graph: str) -> List[dict]:
        """
        Modify a slice
        :param token Fabric Identity Token
        :param slice_id Slice Id
        :param slice_graph Slice Graph Model
        :param ssh_key ssh_key
        :raises Raises an exception in case of failure
        :returns List of reservations created for the Slice on success
        """
        asm_graph = None
        topology = None
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"modify_slice invoked for Controller: {controller}")

            # Check if an Active slice exists already with the same name for the user
            slice_list = controller.get_slices(slice_id=slice_id)
            if slice_list is None or len(slice_list) == 0:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                raise OrchestratorException(f"User# has no Slices",
                                            http_error_code=NOT_FOUND)

            slice_obj = next(iter(slice_list))
            slice_obj.owner.token = token
            if slice_obj.get_graph_id() is None:
                raise OrchestratorException(f"Slice# {slice_obj} does not have graph id")

            slice_state = SliceState(slice_obj.get_state())

            if not SliceState.is_stable(state=slice_state):
                self.logger.info(f"Unable to modify Slice# {slice_id} that is not yet stable, try again later")
                raise OrchestratorException(f"Unable to modify Slice# {slice_id} that is not yet stable, "
                                            f"try again later")

            # Validate the slice graph
            topology = ExperimentTopology(graph_string=slice_graph, importer=NetworkXGraphImporterDisjoint())
            topology.validate()

            asm_graph = FimHelper.get_neo4j_asm_graph(slice_graph=topology.serialize())

            # Authorize the slice
            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.modify, resource=topology)
            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.create, resource=topology)
            project, tags, project_name = fabric_token.first_project
            broker = self.get_broker(controller=controller)
            if broker is None:
                raise OrchestratorException("Unable to determine broker proxy for this controller. "
                                            "Please check Orchestrator container configuration and logs.")

            slice_object = OrchestratorSliceWrapper(controller=controller, broker=broker,
                                                    slice_obj=slice_obj, logger=self.logger)

            # Compute the reservations
            topology_diff, computed_reservations = slice_object.modify(new_slice_graph=asm_graph)
            slice_object.update_topology(topology=topology)

            # Check if Test Bed or site is in maintenance
            self.check_maintenance_mode(token=fabric_token, reservations=computed_reservations)

            # Add any new reservations to the database
            slice_object.add_reservations()

            # Slice has sliver modifications - add/remove/update for slivers requiring AM updates
            modify_state = slice_object.has_sliver_updates_at_authority()
            FimHelper.delete_graph(graph_id=slice_obj.get_graph_id())
            graph_id = asm_graph.get_graph_id()

            slice_obj.graph_id = graph_id
            config_props = slice_obj.get_config_properties()
            config_props[Constants.PROJECT_ID] = project
            config_props[Constants.TAGS] = ','.join(tags)
            config_props[Constants.TOKEN_HASH] = fabric_token.token_hash
            slice_obj.set_config_properties(value=config_props)
            slice_obj.state = SliceState.Modifying.value

            if not controller.update_slice(slice_obj=slice_obj, modify_state=modify_state):
                self.logger.error(f"Failed to update slice: {slice_id} error: {controller.get_last_error()}")

            if modify_state:
                # Enqueue the slice on the demand thread
                # Demand thread is responsible for demanding the reservations
                # Helps improve the create response time
                self.controller_state.get_defer_thread().queue_slice(controller_slice=slice_object)
            # Sliver has meta data update
            else:
                self.logger.debug("Slice only has UserData updates")

            EventLoggerSingleton.get().log_slice_event(slice_object=slice_obj, action=ActionId.modify,
                                                       topology=topology)
            return ResponseBuilder.get_reservation_summary(res_list=computed_reservations)
        except Exception as e:
            if asm_graph is not None:
                FimHelper.delete_graph(graph_id=asm_graph.get_graph_id())

            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing modify_slice e: {e}")
            raise e
        finally:
            if topology is not None and topology.graph_model is not None:
                topology.graph_model.delete_graph()

    def delete_slices(self, *, token: str, slice_id: str = None):
        """
        Delete a user slice identified by slice_id if specified otherwise all user slices within a project
        :param token Fabric Identity Token
        :param slice_id Slice Id
        :raises Raises an exception in case of failure
        """
        try:
            failed_to_delete_slice_ids = []
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"delete_slice invoked for Controller: {controller}")

            slice_guid = ID(uid=slice_id) if slice_id is not None else None
            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.delete)
            project, tags, project_name = fabric_token.first_project

            self.logger.debug(f"Get Slices: {project} {fabric_token.email} {fabric_token.uuid}")
            states = None
            if slice_guid is None:
                states = [SliceState.Nascent.value,
                          SliceState.StableError.value,
                          SliceState.StableOK.value,
                          SliceState.ModifyOK.value,
                          SliceState.ModifyError.value,
                          SliceState.AllocatedError.value,
                          SliceState.AllocatedOK.value]
            slice_list = controller.get_slices(slice_id=slice_guid, user_id=fabric_token.uuid,
                                               project=project, states=states)

            if slice_list is None or len(slice_list) == 0:
                if slice_id is not None:
                    msg = f"Slice# {slice_id} not found"
                    raise OrchestratorException(msg, http_error_code=NOT_FOUND)

            self.__authorize_request(id_token=token, action_id=ActionId.delete)

            for slice_object in slice_list:
                slice_state = SliceState(slice_object.get_state())
                if SliceState.is_dead_or_closing(state=slice_state):
                    self.logger.debug(f"Slice# {slice_object.get_slice_id()} already closed")
                    continue

                if not SliceState.is_stable(state=slice_state) and not SliceState.is_modified(state=slice_state) and \
                        not SliceState.is_allocated(state=slice_state):
                    self.logger.info(f"Unable to delete Slice# {slice_object.get_slice_id()} that is not yet stable, "
                                     f"try again later")
                    failed_to_delete_slice_ids.append(slice_object.get_slice_id())
                    continue

                controller.close_reservations(slice_id=ID(uid=slice_object.get_slice_id()))
            if len(failed_to_delete_slice_ids) > 0:
                raise OrchestratorException(f"Unable to delete Slices {failed_to_delete_slice_ids} that are not yet "
                                            f"stable, try again later")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing delete_slice e: {e}")
            raise e

    def modify_accept(self, *, token: str, slice_id: str) -> dict:
        """
        Accept the last modify on the slice
        :param token Fabric Identity Token
        :param slice_id Slice Id
        :raises Raises an exception in case of failure
        :returns Slice Graph on success
        """
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"modify_accept invoked for Controller: {controller}")

            slice_guid = ID(uid=slice_id) if slice_id is not None else None

            # TODO change this to accept
            self.__authorize_request(id_token=token, action_id=ActionId.modify)

            slice_list = controller.get_slices(slice_id=slice_guid)
            if slice_list is None or len(slice_list) == 0:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                raise OrchestratorException(f"User# has no Slices",
                                            http_error_code=NOT_FOUND)

            slice_obj = next(iter(slice_list))
            slice_state = SliceState(slice_obj.get_state())
            # Do not throw error if modify accept is received for a stable slice
            # Just return the success with slice topology
            #if not SliceState.is_modified(state=slice_state):
            #    self.logger.info(f"Unable to accept modify Slice# {slice_guid} that was not modified")
            #    raise OrchestratorException(f"Unable to accept modify Slice# {slice_guid} that was not modified")

            if slice_obj.get_graph_id() is None:
                raise OrchestratorException(f"Slice# {slice_obj} does not have graph id")

            if not SliceState.is_modified(state=slice_state):
                slice_topology = FimHelper.get_graph(graph_id=slice_obj.get_graph_id())
            # Prune the slice topology only if slice was modified
            else:
                slice_topology = FimHelper.prune_graph(graph_id=slice_obj.get_graph_id())

                controller.accept_update_slice(slice_id=ID(uid=slice_id))

            slice_model_str = slice_topology.serialize()
            return ResponseBuilder.get_slice_summary(slice_list=slice_list, slice_model=slice_model_str)[0]
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing modify_accept e: {e}")
            raise e

    def get_slice_graph(self, *, token: str, slice_id: str, graph_format_str: str, as_self: bool) -> dict:
        """
        Get User Slice
        :param token Fabric Identity Token
        :param slice_id Slice Id
        :param graph_format_str
        :param as_self flag; True - return calling user's slices otherwise, return all slices in the project
        :raises Raises an exception in case of failure
        :returns Slice Graph on success
        """
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"get_slice_graph invoked for Controller: {controller}")

            slice_guid = ID(uid=slice_id) if slice_id is not None else None

            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.query)

            # Filter slices based on user's user_id only when querying as_self
            user_id = fabric_token.uuid
            if not as_self:
                user_id = None

            slice_list = controller.get_slices(slice_id=slice_guid, user_id=user_id)
            if slice_list is None or len(slice_list) == 0:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                raise OrchestratorException(f"User# has no Slices",
                                            http_error_code=NOT_FOUND)

            slice_obj = next(iter(slice_list))

            slice_model_str = None
            if graph_format_str != "NONE":
                if slice_obj.get_graph_id() is None:
                    raise OrchestratorException(f"Slice# {slice_obj} does not have graph id")

                slice_model = FimHelper.get_graph(graph_id=slice_obj.get_graph_id())

                graph_format = self.__translate_graph_format(graph_format=graph_format_str)
                if graph_format == GraphFormat.JSON_NODELINK:
                    slice_model_str = slice_model.serialize_graph()
                    slice_model = FimHelper.get_networkx_graph_from_string(graph_str=slice_model_str)

                if slice_model is None:
                    raise OrchestratorException(f"Slice# {slice_obj} graph could not be loaded")

                slice_model_str = slice_model.serialize_graph(format=graph_format)
            return ResponseBuilder.get_slice_summary(slice_list=slice_list, slice_model=slice_model_str)[0]
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing get_slice_graph e: {e}")
            raise e

    def renew_slice(self, *, token: str, slice_id: str, new_lease_end_time: datetime):
        """
        Renew a slice
        :param token Fabric Identity Token
        :param slice_id Slice Id
        :param new_lease_end_time: New Lease End Time in UTC in '%Y-%m-%d %H:%M:%S %z' format
        :raises Raises an exception in case of failure
        :return:
        """
        failed_to_extend_rid_list = []
        extend_rid_list = []
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"renew_slice invoked for Controller: {controller}")

            slice_guid = ID(uid=slice_id) if slice_id is not None else None
            slice_list = controller.get_slices(slice_id=slice_guid)

            if slice_list is None or len(slice_list) == 0:
                raise OrchestratorException(f"Slice# {slice_id} not found",
                                            http_error_code=NOT_FOUND)

            slice_object = next(iter(slice_list))
            slice_object.owner.token = token

            slice_state = SliceState(slice_object.get_state())
            if SliceState.is_dead_or_closing(state=slice_state):
                raise OrchestratorException(f"Slice# {slice_id} already closed",
                                            http_error_code=BAD_REQUEST)

            if not SliceState.is_stable(state=slice_state) and not SliceState.is_modified(state=slice_state):
                self.logger.info(f"Unable to renew Slice# {slice_guid} that is not yet stable, try again later")
                raise OrchestratorException(f"Unable to renew Slice# {slice_guid} that is not yet stable, "
                                            f"try again later")

            from fabric_cf.actor.security.access_checker import AccessChecker
            fabric_token = AccessChecker.validate_and_decode_token(token=token)
            project, tags, project_name = fabric_token.first_project
            allow_long_lived = True if Constants.SLICE_NO_LIMIT_LIFETIME in tags else False
            _, new_end_time = self.__compute_lease_end_time(lease_end_time=new_lease_end_time,
                                                            allow_long_lived=allow_long_lived,
                                                            project_id=project)

            reservations = controller.get_reservations(slice_id=slice_id)
            if reservations is None or len(reservations) < 1:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                raise OrchestratorException(f"Slice# {slice_id} has no reservations")

            self.logger.debug(f"There are {len(reservations)} reservations in the slice# {slice_id}")

            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.renew,
                                                    lease_end_time=new_end_time)
            self.check_maintenance_mode(token=fabric_token, reservations=reservations)
            for r in reservations:
                res_state = ReservationStates(r.get_state())
                if res_state in [ReservationStates.Closed, ReservationStates.Failed, ReservationStates.CloseWait,
                                 ReservationStates.CloseFail]:
                    continue

                current_end_time = ActorClock.from_milliseconds(milli_seconds=r.get_end())

                if new_end_time < current_end_time:
                    raise OrchestratorException(f"Attempted new term end time is shorter than current slice end time")

                if new_end_time == current_end_time:
                    continue

                self.logger.debug(f"Extending reservation with reservation# {r.get_reservation_id()} "
                                  f"new_end_time: {new_end_time}")
                result = controller.extend_reservation(reservation=ID(uid=r.get_reservation_id()),
                                                       new_end_time=new_end_time,
                                                       sliver=None)
                if not result:
                    self.logger.error(f"Error: {controller.get_last_error()}")
                    failed_to_extend_rid_list.append(r.get_reservation_id())
                else:
                    extend_rid_list.append(r.get_reservation_id())

            '''
            if len(failed_to_extend_rid_list) == 0:
                slice_object.set_lease_end(lease_end=new_end_time)
                if not controller.update_slice(slice_obj=slice_object):
                    self.logger.error(f"Failed to update lease end time: {new_end_time} in Slice: {slice_object}")
                    self.logger.error(controller.get_last_error())
            '''

            if len(failed_to_extend_rid_list) > 0:
                raise OrchestratorException(f"Failed to extend reservation# {failed_to_extend_rid_list}")

            if len(extend_rid_list):
                slice_object.state = SliceState.Configuring.value
                if not controller.update_slice(slice_obj=slice_object, modify_state=True):
                    self.logger.error(f"Failed to update slice: {slice_id} error: {controller.get_last_error()}")

            EventLoggerSingleton.get().log_slice_event(slice_object=slice_object, action=ActionId.renew)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing renew e: {e}")
            raise e

    @staticmethod
    def validate_lease_time(lease_time: str) -> Union[datetime, None]:
        """
        Validate Lease Time
        :param lease_time: Lease Time
        :return Lease Time
        :raises Exception if new lease time is in past
        """
        if lease_time is None:
            return lease_time
        try:
            new_time = datetime.strptime(lease_time, Constants.LEASE_TIME_FORMAT)
        except Exception as e:
            raise OrchestratorException(f"Lease Time is not in format {Constants.LEASE_TIME_FORMAT}",
                                        http_error_code=BAD_REQUEST)

        now = datetime.now(timezone.utc)
        if new_time <= now:
            raise OrchestratorException(f"New lease time {new_time} is in the past! ",
                                        http_error_code=BAD_REQUEST)

        return new_time

    def __compute_lease_end_time(self, lease_end_time: datetime = None, allow_long_lived: bool = False,
                                 project_id: str = None,
                                 lifetime: int = Constants.DEFAULT_LEASE_IN_HOURS) -> tuple[datetime, datetime]:
        """
        Validate and compute Lease End Time.

        :param lease_end_time: The requested end time for the lease.
        :param allow_long_lived: If True, allows extended duration for leases.
        :param project_id: Project ID to check for special duration limits.
        :param lifetime: Requested lease duration in hours. Defaults to the system-defined lease duration.
        :return: A tuple containing the start time (current time) and the computed end time.
        :raises ValueError: If the lease end time is in the past.
        """
        base_time = datetime.now(timezone.utc)

        # Raise an error if lease_end_time is in the past
        if lease_end_time and lease_end_time < base_time:
            raise ValueError("Requested lease end time is in the past.")

        default_max_duration = (Constants.LONG_LIVED_SLICE_TIME_WEEKS
                                if allow_long_lived else Constants.DEFAULT_MAX_DURATION_IN_WEEKS).total_seconds()
        # Convert weeks to hours
        default_max_duration /= 3600

        # Calculate lifetime if not directly provided
        if lease_end_time:
            lifetime = (lease_end_time - base_time).total_seconds() / 3600
        else:
            if not lifetime:
                lifetime = Constants.DEFAULT_LEASE_IN_HOURS

        # Ensure the requested lifetime does not exceed allowed max duration for the project
        if project_id not in self.infrastructure_project_id and lifetime > default_max_duration:
            self.logger.info(f"Requested lifetime ({lifetime} hours) exceeds the allowed duration "
                             f"({default_max_duration} hours). Setting to maximum allowable.")
            lifetime = default_max_duration

        # Calculate the new end time
        new_end_time = base_time + timedelta(hours=lifetime)

        return base_time, new_end_time

    @staticmethod
    def __translate_graph_format(*, graph_format: str) -> GraphFormat:
        if graph_format == GraphFormat.GRAPHML.name:
            return GraphFormat.GRAPHML
        elif graph_format == GraphFormat.JSON_NODELINK.name:
            return GraphFormat.JSON_NODELINK
        elif graph_format == GraphFormat.CYTOSCAPE.name:
            return GraphFormat.CYTOSCAPE
        else:
            return GraphFormat.GRAPHML

    def check_maintenance_mode(self, *, token: FabricToken, reservations: List[ReservationMng] = None):
        controller = self.controller_state.get_management_actor()
        self.logger.debug(f"check_maintenance_mode invoked for Controller: {controller}")

        project, tags, project_name = token.first_project

        if not controller.is_slice_provisioning_allowed(project=project, email=token.email):
            raise OrchestratorException(Constants.MAINTENANCE_MODE_ERROR,
                                        http_error_code=Constants.INTERNAL_SERVER_ERROR_MAINT_MODE)

        if reservations is not None:
            for r in reservations:
                sliver = r.get_sliver()
                if not isinstance(sliver, NetworkServiceSliver):
                    worker = None
                    if sliver.get_labels() is not None and sliver.get_labels().instance_parent is not None:
                        worker = sliver.get_labels().instance_parent
                    status, message = controller.is_sliver_provisioning_allowed(project=project,
                                                                                site=sliver.get_site(),
                                                                                email=token.email,
                                                                                worker=worker)
                    if not status:
                        raise OrchestratorException(message=message,
                                                    http_error_code=Constants.INTERNAL_SERVER_ERROR_MAINT_MODE)

    def poa(self, *, token: str, sliver_id: str, poa: PoaAvro) -> tuple[str, str]:
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"poa invoked for Controller: {controller} pos: {poa}")

            rid = ID(uid=sliver_id) if sliver_id is not None else None

            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.POA)
            user_id = fabric_token.uuid
            project, tags, project_name = fabric_token.first_project

            auth = AuthAvro()
            auth.name = self.controller_state.get_management_actor().get_name()
            auth.guid = self.controller_state.get_management_actor().get_guid()
            auth.oidc_sub_claim = fabric_token.uuid
            auth.email = fabric_token.email
            poa.auth = auth
            poa.project_id = project
            poa.rid = sliver_id

            reservations = controller.get_reservations(rid=rid, oidc_claim_sub=user_id)
            if reservations is None or len(reservations) != 1:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                    if controller.get_last_error().status.code == ErrorCodes.ErrorNoSuchReservation:
                        raise OrchestratorException(f"Reservation# {rid} not found",
                                                    http_error_code=NOT_FOUND)

                raise OrchestratorException(f"Reservation# {rid} not found",
                                            http_error_code=NOT_FOUND)

            res_state = ReservationStates(reservations[0].get_state())

            if res_state != ReservationStates.Active:
                raise OrchestratorException(f"Cannot trigger POA; Reservation# {rid} is not {ReservationStates.Active}")

            if not controller.poa(poa=poa):
                raise OrchestratorException(f"Failed to trigger POA: "
                                            f"{controller.get_last_error().get_status().get_message()}")
            self.logger.debug(f"POA {poa.operation}/{sliver_id} added successfully")
            return poa.poa_id, reservations[0].get_slice_id()
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing poa e: {e}")
            raise e

    def get_poas(self, *, token: str, sliver_id: str = None, poa_id: str = None, states: List[str] = None,
                 limit: int = 200, offset: int = 0):
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"poa invoked for Controller: {controller}")

            rid = ID(uid=sliver_id) if sliver_id is not None else None

            fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.query)
            email = fabric_token.email
            project, tags, project_name = fabric_token.first_project

            poa_states = PoaStates.translate_list(states=states)

            auth = AuthAvro()
            auth.name = self.controller_state.get_management_actor().get_name()
            auth.guid = self.controller_state.get_management_actor().get_guid()
            auth.oidc_sub_claim = fabric_token.uuid
            auth.email = fabric_token.email

            poa_list = controller.get_poas(rid=rid, poa_id=poa_id, project_id=project,
                                           states=poa_states, limit=limit, offset=offset)
            if poa_list is None:
                if controller.get_last_error() is not None:
                    self.logger.error(controller.get_last_error())
                    if controller.get_last_error().status.code == ErrorCodes.ErrorNoSuchPoa:
                        raise OrchestratorException(f"Reservation# {rid} not found",
                                                    http_error_code=NOT_FOUND)

                raise OrchestratorException(f"{controller.get_last_error()}")
            return ResponseBuilder.get_poa_summary(poa_list=poa_list)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing poa e: {e}")
            raise e

    def get_metrics_overview(self, *, token: str = None, excluded_projects: List[str] = None):
        """
        Get metrics overview
        :param token: token
        :param excluded_projects: list of projects to exclude
        """
        try:
            controller = self.controller_state.get_management_actor()
            self.logger.debug(f"get_metrics_overview invoked for Controller: {controller}")

            project = None
            user_id = None
            # Filter based on project_id and user_id when token is provided
            if token:
                fabric_token = self.__authorize_request(id_token=token, action_id=ActionId.query)
                projects = fabric_token.projects
                if len(projects) == 1:
                    project, tags, project_name = fabric_token.first_project
                user_id = fabric_token.uuid

            if excluded_projects:
                excluded_projects.extend(self.excluded_projects)
            else:
                excluded_projects = self.excluded_projects

            active_states = SliceState.list_values_ex_closing_dead()
            active_slice_count = controller.get_slice_count(states=active_states, user_id=user_id, project=project,
                                                            excluded_projects=excluded_projects)
            non_active_metrics = controller.get_metrics(oidc_sub=user_id, project_id=project,
                                                        excluded_projects=excluded_projects)
            total_slices = 0
            for m in non_active_metrics:
                total_slices += m.get("slice_count", 0)
            if not user_id and not project:
                # Get Seed value from config
                total_slices += self.total_slice_count_seed
            result = {
                "slices": {
                    "active_cumulative": active_slice_count,
                    "non_active_cumulative": total_slices
                }
            }
            return result
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(f"Exception occurred processing get_metrics_overview e: {e}")
            raise e



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime, timezone

from fabric_cf.actor.fim.fim_helper import FimHelper
from fim.user import GraphFormat


class BqmWrapper:
    """
    Implements cache for storing the BQM
    """
    def __init__(self):
        self.graph_format = None
        self.bqm = None
        self.last_query_time = None
        self.refresh_interval_in_seconds = 60
        self.refresh_in_progress = False
        self.level = 1
        self.graph_id = None

    def set_refresh_interval(self, *, refresh_interval: int):
        """
        Set BQM Refresh Interval
        @param refresh_interval: Refresh Interval in seconds
        """
        self.refresh_interval_in_seconds = refresh_interval

    def can_refresh(self) -> bool:
        """
        Determine if it's time to refresh the BQM
        @return True -> On first attempt or last query time is after refresh Interval; False otherwise
        """
        current_time = datetime.now(timezone.utc)
        if not self.refresh_in_progress and (self.last_query_time is None or
                                             ((current_time - self.last_query_time).seconds >
                                              self.refresh_interval_in_seconds)):
            return True
        return False

    def save(self, *, bqm: str, graph_format: GraphFormat, level: int):
        """
        Save the BQM
        @param bqm: Broker Query Model
        @param graph_format: Broker Query Model Format
        @param level: Level
        """
        self.graph_format = graph_format
        self.bqm = bqm
        self.last_query_time = datetime.now(timezone.utc)
        self.refresh_in_progress = False
        self.level = level

    def start_refresh(self):
        self.refresh_in_progress = True

    def get_bqm(self) -> str:
        """
        Get BQM
        @return returns BQM
        """
        return self.bqm

    def get_graph_format(self) -> GraphFormat:
        """
        Get BQM Format
        @return bqm format
        """
        return self.graph_format

    def get_level(self) -> int:
        """
        Return BQM level
        """
        return self.level



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING, List

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.util.reflection_utils import ReflectionUtils
from fabric_cf.actor.core.manage.management_object import ManagementObject

if TYPE_CHECKING:
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.apis.abc_container_database import ABCContainerDatabase
    from fabric_cf.actor.core.apis.abc_management_object import ABCManagementObject


class ManagementObjectManager:
    """
    This class makes it possible to register, index, and persist information about ManagementObjects.
    A ManagementObject can be registered only if no other object with the same identifier has been registered.
    Registering the same object more than once is not permitted.
    Each successfully registered management object is serialized and stored in the database.
    Note that once stored, this object cannot be updated. The stored information is sufficient to recreate an instance
    of the ManagementObject. Objects deriving from ManagementObject are responsible for their own persistence.
    """

    def __init__(self):
        # "Live" manager objects. These are all manager objects that have been instantiated inside this container.
        self.objects = {}
        # The database.
        self.db = None
        # logger
        self.logger = None
        # Initialization status.
        self.initialized = False
        self.lock = threading.Lock()

    def create_instance(self, *, mo_obj: dict) -> ManagementObject:
        if Constants.PROPERTY_CLASS_NAME not in mo_obj or Constants.PROPERTY_MODULE_NAME not in mo_obj:
            raise ManageException("Missing class name")

        class_name = mo_obj[Constants.PROPERTY_CLASS_NAME]
        module_name = mo_obj[Constants.PROPERTY_MODULE_NAME]

        self.logger.debug("Creating management object: {}".format(class_name))

        obj = ReflectionUtils.create_instance(module_name=module_name, class_name=class_name)

        if not isinstance(obj, ManagementObject):
            raise ManageException("Object does not implement ManagementObject interface")

        obj.reset(properties=mo_obj)
        obj.initialize()

        return obj

    def get_management_object(self, *, key: ID):
        """
        Retrieves the specified manager object.
        @param key object guid
        @return returns Management object
        """
        try:
            self.logger.debug("key: {} object: {}".format(key, self.objects.get(key, None)))
            self.lock.acquire()
            return self.objects.get(key, None)
        finally:
            self.lock.release()

    def initialize(self, *, db: ABCContainerDatabase):
        """
        Performs initialization. If the system is recovering after a
        shutdown/crash, loads all manager objects that pertain to the container.
        @param db database
        @throws Exception in case of error
        """
        if db is None:
            raise ManageException("database cannot be null")

        if not self.initialized:
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            self.logger = GlobalsSingleton.get().get_logger()
            self.db = db
            if not GlobalsSingleton.get().get_container().is_fresh():
                self.load_container_manager_objects()
            self.initialized = True

    def load_actor_manager_objects(self, *, actor_name: str):
        """
        Loads all manager objects associated with the specific actor.
        @param actor_name actor name
        @throws Exception in case of error
        """
        self.logger.info("Loading container-level management objects for actor: {}".format(actor_name))
        manager_objects = self.db.get_manager_objects_by_actor_name(actor_name=actor_name)
        self.load_objects(manager_objects=manager_objects)
        self.logger.info("Finished loading container-level management objects for actor: {}".format(actor_name))

    def load_container_manager_objects(self):
        """
        Loads all manager objects not associated with specific actors
        @throws Exception in case of error
        """
        self.logger.info("Loading container-level management objects")
        manager_objects = self.db.get_manager_container()
        if manager_objects is None:
            return
        self.load_objects(manager_objects=manager_objects)
        self.logger.info("Finished loading container-level management objects")

    def load_objects(self, *, manager_objects: List[dict]):
        """
        Loads the specified management objects
        @param manager_objects list of properties
        @throws Exception in case of error
        """
        if manager_objects is None:
            return
        for m in manager_objects:
            manager = self.create_instance(mo_obj=m[Constants.PROPERTY_PICKLE_PROPERTIES])
            try:
                self.lock.acquire()
                if manager.get_id() in self.objects:
                    raise ManageException("there is already a management object in memory with the specified id")

                self.objects[manager.get_id()] = manager

            finally:
                self.lock.release()

            self.logger.info("Loaded management object with id: {}".format(manager.get_id()))

    def register_manager_object(self, *, manager: ABCManagementObject):
        self.db.add_manager_object(manager=manager)
        try:
            self.lock.acquire()
            self.objects[manager.get_id()] = manager
        finally:
            self.lock.release()

    def unload_actor_manager_objects(self, *, actor_name: str):
        manager_objects = self.db.get_manager_objects_by_actor_name(actor_name=actor_name)
        self.unload_objects(manager_objects=manager_objects)

    def unload_objects(self, *, manager_objects: list):
        if manager_objects is not None:
            for manager in manager_objects:
                mgr_id = manager['mo_key']
                try:
                    self.lock.acquire()
                    if mgr_id in self.objects:
                        self.objects.pop(mgr_id)
                finally:
                    self.lock.release()

    def unregister_manager_object(self, *, mid: ID):
        try:
            self.lock.acquire()
            if mid in self.objects:
                self.objects.pop(mid)
        finally:
            self.lock.release()

            self.db.remove_manager_object(manager_key=mid)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Ilya Baldin (ibaldin@renci.org)
from __future__ import annotations

import json
from datetime import datetime
from typing import Tuple, Dict, TYPE_CHECKING, List
from collections import defaultdict

import uuid

from fim.graph.abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
from fim.graph.resources.abc_cbm import ABCCBMPropertyGraph
from fim.graph.resources.abc_bqm import ABCBQMPropertyGraph
from fim.graph.networkx_property_graph import NetworkXGraphImporter
from fim.graph.resources.networkx_abqm import NetworkXAggregateBQM
from fim.slivers.capacities_labels import Capacities, Flags, Labels
from fim.slivers.delegations import DelegationFormat
from fim.slivers.maintenance_mode import MaintenanceInfo, MaintenanceEntry, MaintenanceState
from fim.slivers.network_node import CompositeNodeSliver, NodeType, NodeSliver
from fim.slivers.attached_components import ComponentSliver, ComponentType, AttachedComponentsInfo
from fim.slivers.interface_info import InterfaceType
from fim.slivers.network_service import ServiceType

from fabric_cf.actor.core.kernel.reservation_states import ReservationStates

if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_database import ABCDatabase


class AggregatedBQMPlugin:
    """
    Implement a plugin for simple aggregation of CBM into BQM, transforming site
    topologies into CompositeNodes and linking them with (Composite)Links
    This is based on fim.pluggable BrokerPluggable
    """
    # set to true to test creating ABQM without talking to database or requiring
    # an actor reference
    DEBUG_FLAG = False

    def __init__(self, actor, logger=None):
        if not self.DEBUG_FLAG:
            assert actor is not None
        self.actor = actor
        self.logger = logger

    @staticmethod
    def _remove_none_entries(d):
        return {k: v for (k, v) in d.items() if v}

    def __site_maintenance_info(self, *, site_name: str):
        if self.actor is None:
            return None
        site = self.actor.get_plugin().get_database().get_site(site_name=site_name)
        if site is not None:
            result = site.get_maintenance_info().copy()
            if result.get(site_name) is None:
                entry = MaintenanceEntry(state=MaintenanceState.Active)
                result.add(site_name, entry)
        else:
            result = MaintenanceInfo()
            entry = MaintenanceEntry(state=MaintenanceState.Active)
            result.add(site_name, entry)
        result.finalize()
        return result

    @staticmethod
    def occupied_vlans(db: ABCDatabase, node_id: str, component_name: str, start: datetime = None,
                       end: datetime = None) -> List[str]:
        """
        Get existing components attached to Active/Ticketed Network Service Slivers
        :param db:
        :param node_id:
        :param component_name:
        :param start:
        :param end:
        :return: list of components
        """
        assert node_id is not None
        states = [ReservationStates.Active.value,
                  ReservationStates.ActiveTicketed.value,
                  ReservationStates.Ticketed.value,
                  ReservationStates.Nascent.value]

        result = []
        res_type = []
        for x in ServiceType:
            res_type.append(str(x))

        # Only get Active or Ticketing reservations
        comps = db.get_components(node_id=node_id, component=component_name, rsv_type=res_type, states=states,
                                  start=start, end=end)
        if comps is not None:
            if comps.get(component_name):
                result = comps.get(component_name)
        return result

    @staticmethod
    def occupied_link_capacity(*, db: ABCDatabase, node_id: str, start: datetime, end: datetime) -> Capacities:
        """
        Compute the total bandwidth capacity occupied on a given link node within a specific time window.

        This method queries the database for reservations on the specified link node that are in active or
        ticketing states and fall within the given start and end time window. It sums up the capacity used
        by all such relevant reservations.

        :param db: An instance of ABCDatabase used to query reservations.
        :param node_id: The unique identifier of the link node to check.
        :param start: The start time of the reservation window.
        :param end: The end time of the reservation window.
        :return: Total Capacities in dict containing the occupied bandwidth (in Gbps) on the link during the
        specified time window.
        """

        states = [ReservationStates.Active.value,
                  ReservationStates.ActiveTicketed.value,
                  ReservationStates.Ticketed.value,
                  ReservationStates.Nascent.value]

        res_type = []
        for x in ServiceType:
            res_type.append(str(x))

        # Only get Active or Ticketing reservations
        existing = db.get_links(node_id=node_id, rsv_type=res_type, states=states, start=start, end=end)

        bw_used = existing.get(node_id, 0)
        if bw_used:
            return Capacities(bw=bw_used).to_dict()

    @staticmethod
    def occupied_node_capacity(*, db: ABCDatabase, node_id: str, start: datetime,
                               end: datetime) -> Tuple[Capacities, Dict[ComponentType, Dict[str, Capacities]]]:
        """
        Figure out the total capacity occupied in the network node and return a tuple of
        capacities occupied in this node and a dict of component capacities that are occupied
        organized by component type and model.
        """
        assert node_id is not None
        states = [ReservationStates.Active.value,
                  ReservationStates.ActiveTicketed.value,
                  ReservationStates.Ticketed.value,
                  ReservationStates.Nascent.value]

        # get existing reservations for this node
        existing_reservations = db.get_reservations(graph_node_id=node_id, states=states, start=start, end=end)
        # node capacities
        occupied_capacities = Capacities()
        occupied_component_capacities = defaultdict(dict)
        # Remove allocated capacities to the reservations
        if existing_reservations is not None:
            for reservation in existing_reservations:
                # For Active or Ticketed or Ticketing reservations; compute the counts from available
                allocated_sliver = None
                if reservation.is_ticketing() and reservation.get_approved_resources() is not None:
                    allocated_sliver = reservation.get_approved_resources().get_sliver()

                if (reservation.is_active() or reservation.is_ticketed()) and \
                        reservation.get_resources() is not None:
                    allocated_sliver = reservation.get_resources().get_sliver()

                if allocated_sliver is not None:
                    occupied_capacities = occupied_capacities + allocated_sliver.get_capacity_allocations()

                    if allocated_sliver.attached_components_info is not None:
                        for allocated_component in allocated_sliver.attached_components_info.devices.values():
                            # Ignore components without any allocations
                            if not allocated_component.capacity_allocations:
                                continue
                            rt = allocated_component.resource_type
                            rm = allocated_component.resource_model
                            if occupied_component_capacities[rt].get(rm) is None:
                                occupied_component_capacities[rt][rm] = Capacities()

                            occupied_component_capacities[rt][rm] = occupied_component_capacities[rt][rm] + \
                                                          allocated_component.capacity_allocations

        return occupied_capacities, occupied_component_capacities

    def plug_produce_bqm(self, *, cbm: ABCCBMPropertyGraph, **kwargs) -> ABCBQMPropertyGraph:
        """
        Generate a site-aggregated BQM (Broker Query Model) graph from a CBM (Combined Broker Model).

        This function aggregates network and compute resources per site, organizing them based on the
        specified query level. The resulting graph is suitable for advertisement, resource lookup, and
        scheduling decisions.

        Query Levels:
            - Level 0: Includes full per-worker detail with original node IDs (used by Orchestrator for advanced scheduling)
            - Level 1: Includes consolidated per-site information only, with anonymized node IDs (used by `fablib.get_sites()`)
            - Level 2: Includes per-worker detail with anonymized node IDs (default for `fablib.get_sites()`)

        Parameters:
            cbm (ABCCBMPropertyGraph): Input graph containing raw broker data including nodes, links, and components.
            **kwargs:
                query_level (int): Level of aggregation (0, 1, or 2). Defaults to Level 2 if unspecified.
                includes (str): Comma-separated list of site names to include.
                excludes (str): Comma-separated list of site names to exclude.
                start (datetime): Optional start time for querying capacity allocations.
                end (datetime): Optional end time for querying capacity allocations.

        Returns:
            ABCBQMPropertyGraph: An aggregated NetworkX-based BQM graph reflecting the requested query level.
        """
        if kwargs.get('query_level', None) is None or kwargs['query_level'] > 2:
            return cbm.clone_graph(new_graph_id=str(uuid.uuid4()))

        includes = kwargs.get('includes', None)
        excludes = kwargs.get('excludes', None)

        sites_to_include = [s.strip().upper() for s in includes.split(",")] if includes else []
        sites_to_exclude = [s.strip().upper() for s in excludes.split(",")] if excludes else []

        start = kwargs.get('start', None)
        end = kwargs.get('end', None)
        if not self.DEBUG_FLAG:
            db = self.actor.get_plugin().get_database()
        else:
            db = None

        # do a one-pass aggregation of servers, their components and interfaces
        # and some flags (e.g. PTP availability)
        # this includes facilities
        nnodes = cbm.get_all_nodes_by_class(label=ABCPropertyGraph.CLASS_NetworkNode)
        slivers_by_site = defaultdict(list)
        p4s_by_site = defaultdict(list)
        for n in nnodes:
            # build deep slivers for each advertised server, aggregate by site
            node_sliver = cbm.build_deep_node_sliver(node_id=n)
            slivers_by_site[node_sliver.site].append(node_sliver)
            if node_sliver.get_type() == NodeType.Switch and "p4" in node_sliver.get_name():
                p4s_by_site[node_sliver.site].append(node_sliver)

        # create a new blank Aggregated BQM NetworkX graph
        if kwargs['query_level'] == 0:
            graph_id = cbm.graph_id
            if kwargs.get('graph_id'):
                graph_id = kwargs.get('graph_id')
            abqm = NetworkXAggregateBQM(graph_id=graph_id,
                                        importer=NetworkXGraphImporter(logger=self.logger),
                                        logger=self.logger)
        else:
            abqm = NetworkXAggregateBQM(graph_id=str(uuid.uuid4()),
                                        importer=NetworkXGraphImporter(logger=self.logger),
                                        logger=self.logger)

        site_to_composite_node_id = dict()
        site_to_ns_node_id = dict()
        facilities_by_site = defaultdict(list)
        for s, ls in slivers_by_site.items():
            if len(sites_to_include) and s not in sites_to_include:
                continue
            if len(sites_to_exclude) and s in sites_to_exclude:
                continue
            # add up capacities and delegated capacities, skip labels for now
            # count up components and figure out links between site

            site_sliver = CompositeNodeSliver()
            # count what is taken
            site_sliver.capacity_allocations = Capacities()
            # count what is available
            site_sliver.capacities = Capacities()
            site_sliver.resource_name = s
            site_sliver.resource_type = NodeType.Server
            site_sliver.node_id = str(uuid.uuid4())
            # available components organized by [type][model]
            site_comps_by_type = defaultdict(dict)
            # occupied component capacities organized by [type][model] into lists (by server)
            site_allocated_comps_caps_by_type = defaultdict(dict)
            site_sliver.maintenance_info = self.__site_maintenance_info(site_name=s)

            loc = None
            ptp = False
            workers = []
            for sliver in ls:
                if sliver.get_type() != NodeType.Server:
                    # skipping NAS, Facility and dataplane switches
                    if sliver.get_type() == NodeType.Facility:
                        # keep track of facilities for each site
                        facilities_by_site[s].append(sliver)
                    continue
                # Build worker sliver to hold per worker information
                worker_sliver = NodeSliver()
                worker_sliver.capacities = Capacities()
                worker_sliver.capacity_allocations = Capacities()
                worker_sliver.resource_name = sliver.get_name()
                worker_sliver.resource_type = NodeType.Server
                worker_sliver.set_site(s)
                worker_sliver.node_id = str(uuid.uuid4())
                if self.DEBUG_FLAG or kwargs['query_level'] == 0:
                    # for debugging and running in a test environment
                    # also for level 0; only return capacity information
                    allocated_comp_caps = dict()
                    worker_sliver.node_id = sliver.node_id
                else:
                    # query database for everything taken on this node
                    allocated_caps, allocated_comp_caps = self.occupied_node_capacity(db=db, node_id=sliver.node_id,
                                                                                      start=start, end=end)
                    site_sliver.capacity_allocations = site_sliver.capacity_allocations + allocated_caps
                    worker_sliver.capacity_allocations = allocated_caps

                # get the location if available
                if loc is None:
                    loc = sliver.get_location()
                worker_sliver.set_location(loc)

                # look at flags
                flags = sliver.get_flags()
                if flags and not ptp and flags.ptp:
                    ptp = True

                site_sliver.set_flags(Flags(ptp=ptp))

                # calculate available node capacities based on delegations
                if sliver.get_capacity_delegations() is not None:
                    # CBM only has one delegation if it has one
                    _, delegation = sliver.get_capacity_delegations().get_sole_delegation()
                    # FIXME: skip pool definitions and references for now
                    if delegation.get_format() == DelegationFormat.SinglePool:
                        site_sliver.capacities = site_sliver.capacities + \
                            delegation.get_details()
                        worker_sliver.capacities = delegation.get_details()
                # This for the case when BQM is generated from Orchestrator
                else:
                    site_sliver.capacities += sliver.get_capacities()
                    worker_sliver.capacities = sliver.get_capacities()

                    # collect available components in lists by type and model for the site (for later aggregation)
                if sliver.attached_components_info is None:
                    continue
                worker_sliver.attached_components_info = AttachedComponentsInfo()
                for comp in sliver.attached_components_info.list_devices():
                    rt = comp.resource_type
                    rm = comp.resource_model
                    if site_comps_by_type[rt].get(rm) is None:
                        site_comps_by_type[rt][rm] = list()
                    site_comps_by_type[rt][rm].append(comp)
                    name = f"{rt}-{rm}"
                    # Add per worker component availability
                    if worker_sliver.attached_components_info.devices.get(name) is None:
                        worker_comp_sliver = ComponentSliver()
                        worker_comp_sliver.resource_name = name
                        worker_comp_sliver.capacities = Capacities()
                        worker_comp_sliver.capacity_allocations = Capacities()
                        worker_comp_sliver.node_id = str(uuid.uuid4())
                        worker_comp_sliver.set_type(rt)
                        worker_comp_sliver.set_model(rm)
                        worker_sliver.attached_components_info.devices[name] = worker_comp_sliver

                    worker_sliver.attached_components_info.devices[name].capacities += comp.capacities

                # merge allocated component capacities
                for kt, v in allocated_comp_caps.items():
                    for km, vcap in v.items():
                        if site_allocated_comps_caps_by_type[kt].get(km) is None:
                            site_allocated_comps_caps_by_type[kt][km] = Capacities()
                        site_allocated_comps_caps_by_type[kt][km] = site_allocated_comps_caps_by_type[kt][km] + \
                                                                    vcap
                        name = f"{kt}-{km}"
                        if worker_sliver.attached_components_info.devices.get(name) is not None:
                            worker_sliver.attached_components_info.devices[name].capacity_allocations += vcap

                workers.append(worker_sliver)

            # set location to whatever is available
            site_sliver.set_location(loc)
            site_sliver.set_site(s)
            site_sliver.set_flags(Flags(ptp=ptp))

            # create a Composite node for every site
            site_to_composite_node_id[s] = site_sliver.node_id
            site_props = abqm.node_sliver_to_graph_properties_dict(site_sliver)
            abqm.add_node(node_id=site_sliver.node_id, label=ABCPropertyGraph.CLASS_CompositeNode,
                          props=site_props)

            p4s = p4s_by_site.get(site_sliver.site)
            if p4s:
                for p4 in p4s:
                    p4_sliver = NodeSliver()
                    p4_sliver.node_id = str(uuid.uuid4())
                    p4_sliver.set_type(resource_type=NodeType.Switch)
                    p4_sliver.set_name(resource_name=p4.get_name())
                    p4_sliver.set_site(site=p4.get_site())
                    p4_sliver.capacities = Capacities()
                    p4_sliver.capacity_allocations = Capacities()
                    p4_sliver.capacities += p4.get_capacities()
                    if not self.DEBUG_FLAG:
                        # query database for everything taken on this node
                        allocated_caps, allocated_comp_caps = self.occupied_node_capacity(db=db, node_id=p4.node_id,
                                                                                          start=start, end=end)
                        if allocated_caps:
                            p4_sliver.capacity_allocations = p4_sliver.capacity_allocations + allocated_caps
                    p4_props = abqm.node_sliver_to_graph_properties_dict(p4_sliver)
                    node_id = p4.node_id
                    if kwargs['query_level'] != 0:
                        node_id = p4_sliver.node_id
                    abqm.add_node(node_id=node_id, label=ABCPropertyGraph.CLASS_NetworkNode, props=p4_props)
                    abqm.add_link(node_a=site_sliver.node_id, rel=ABCPropertyGraph.REL_HAS, node_b=node_id)

            # Add per worker metrics for query level 2
            if kwargs['query_level'] == 2 or kwargs['query_level'] == 0:
                for w in workers:
                    # Add workers
                    abqm.add_node(node_id=w.node_id, label=ABCPropertyGraph.CLASS_NetworkNode,
                                  props=abqm.node_sliver_to_graph_properties_dict(w))
                    # Connect workers to Site Node
                    abqm.add_link(node_a=site_sliver.node_id, rel=ABCPropertyGraph.REL_HAS, node_b=w.node_id)
                    if w.attached_components_info is None:
                        continue
                    # Add components
                    for c in w.attached_components_info.list_devices():
                        abqm.add_node(node_id=c.node_id, label=ABCPropertyGraph.CLASS_Component,
                                      props=abqm.node_sliver_to_graph_properties_dict(c))
                        # Connect components to the worker
                        abqm.add_link(node_a=w.node_id, rel=ABCPropertyGraph.REL_HAS, node_b=c.node_id)

            # add a network service
            ns_id = str(uuid.uuid4())
            site_to_ns_node_id[s] = ns_id
            ns_props = {ABCPropertyGraph.PROP_NAME: s + '_ns',
                        ABCPropertyGraph.PROP_TYPE: str(ServiceType.MPLS)}
            abqm.add_node(node_id=ns_id, label=ABCPropertyGraph.CLASS_NetworkService, props=ns_props)
            abqm.add_link(node_a=site_sliver.node_id, rel=ABCPropertyGraph.REL_HAS, node_b=ns_id)

            # Site level component information is only passed for query_level=1
            if kwargs['query_level'] == 1:
                # create a component sliver for every component type/model pairing
                # and add a node for it linking back to site node
                for ctype, cdict in site_comps_by_type.items():
                    for cmodel, comp_list in cdict.items():
                        comp_sliver = ComponentSliver()
                        # count what is available
                        comp_sliver.capacities = Capacities()
                        # count what is taken (ignore those type/model pairings that were unused)
                        comp_sliver.capacity_allocations = site_allocated_comps_caps_by_type[ctype].get(cmodel) or \
                                                           Capacities()
                        comp_sliver.set_type(ctype)
                        comp_sliver.set_model(cmodel)
                        comp_sliver.set_name(str(ctype) + '-' + cmodel)
                        for comp in comp_list:
                            comp_sliver.capacities = comp_sliver.capacities + comp.capacities
                        comp_node_id = str(uuid.uuid4())
                        comp_props = abqm.component_sliver_to_graph_properties_dict(comp_sliver)
                        abqm.add_node(node_id=comp_node_id, label=ABCPropertyGraph.CLASS_Component,
                                      props=comp_props)
                        abqm.add_link(node_a=site_sliver.node_id, rel=ABCPropertyGraph.REL_HAS,
                                      node_b=comp_node_id)

        # get all intersite links - add them to the aggregated BQM graph
        intersite_links = cbm.get_intersite_links()
        for l in intersite_links:
            source_switch = l[0]
            sink_switch = l[2]
            link = l[1]
            source_site = l[3]
            sink_site = l[4]
            source_cp = l[5]
            sink_cp = l[6]
            # Exclude the sites requested to be filtered
            if includes and (source_site not in includes or sink_site not in includes):
                continue
            if excludes and (source_site in excludes or sink_site not in excludes):
                continue
            _, cbm_source_cp_props = cbm.get_node_properties(node_id=source_cp)
            _, cbm_sink_cp_props = cbm.get_node_properties(node_id=sink_cp)
            _, cbm_link_props = cbm.get_node_properties(node_id=link)
            # add connection point, link, connection point between two NetworkServices
            assert(site_to_ns_node_id.get(source_site) is not None and
                   site_to_ns_node_id.get(sink_site) is not None)
            source_cp_id = str(uuid.uuid4())
            sink_cp_id = str(uuid.uuid4())
            source_cp_props = {ABCPropertyGraph.PROP_NAME: "_".join([source_site, sink_site]),
                               ABCPropertyGraph.PROP_TYPE: str(InterfaceType.TrunkPort),
                               ABCPropertyGraph.PROP_CLASS: ABCPropertyGraph.CLASS_ConnectionPoint,
                               ABCPropertyGraph.PROP_LABELS: cbm_source_cp_props.get(ABCPropertyGraph.PROP_LABELS),
                               ABCPropertyGraph.PROP_CAPACITIES: cbm_source_cp_props.get(ABCPropertyGraph.PROP_CAPACITIES)
                               }
            source_cp_props = {k: v for (k, v) in source_cp_props.items() if v}

            abqm.add_node(node_id=source_cp_id, label=ABCPropertyGraph.CLASS_ConnectionPoint,
                          props=source_cp_props)
            # FIXME: CP names may not be unique if we are dealing with a multigraph
            sink_cp_props = {ABCPropertyGraph.PROP_NAME: "_".join([sink_site, source_site]),
                             ABCPropertyGraph.PROP_TYPE: str(InterfaceType.TrunkPort),
                             ABCPropertyGraph.PROP_CLASS: ABCPropertyGraph.CLASS_ConnectionPoint,
                             ABCPropertyGraph.PROP_LABELS: cbm_sink_cp_props.get(ABCPropertyGraph.PROP_LABELS),
                             ABCPropertyGraph.PROP_CAPACITIES: cbm_sink_cp_props.get(ABCPropertyGraph.PROP_CAPACITIES)
                             }
            sink_cp_props = {k: v for (k, v) in sink_cp_props.items() if v}
            abqm.add_node(node_id=sink_cp_id, label=ABCPropertyGraph.CLASS_ConnectionPoint,
                          props=sink_cp_props)
            # selectively replicate link node and its properties from CBM
            new_link_props = {ABCPropertyGraph.PROP_NAME: cbm_link_props[ABCPropertyGraph.PROP_NAME],
                              ABCPropertyGraph.PROP_TYPE: cbm_link_props[ABCPropertyGraph.PROP_TYPE],
                              ABCPropertyGraph.PROP_CLASS: cbm_link_props[ABCPropertyGraph.PROP_CLASS],
                              ABCPropertyGraph.PROP_LAYER: cbm_link_props[ABCPropertyGraph.PROP_LAYER],
                              }
            if kwargs['query_level'] == 0:
                if cbm_link_props.get(ABCPropertyGraph.PROP_CAPACITIES):
                    new_link_props[ABCPropertyGraph.PROP_CAPACITIES] = cbm_link_props[ABCPropertyGraph.PROP_CAPACITIES]
                if cbm_link_props.get(ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS):
                    new_link_props[ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS] = cbm_link_props[ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS]
            else:
                if cbm_link_props.get(ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS):
                    new_link_props[ABCPropertyGraph.PROP_CAPACITIES] = cbm_link_props[
                        ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS]
                elif cbm_link_props.get(ABCPropertyGraph.PROP_CAPACITIES):
                    new_link_props[ABCPropertyGraph.PROP_CAPACITIES] = cbm_link_props[ABCPropertyGraph.PROP_CAPACITIES]
                if not self.DEBUG_FLAG:
                    occupied_link_capacity = self.occupied_link_capacity(node_id=link, db=db, start=start, end=end)
                    if occupied_link_capacity:
                        new_link_props[ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS] = occupied_link_capacity

            abqm.add_node(node_id=link, label=ABCPropertyGraph.CLASS_Link, props=new_link_props)
            # connect them together
            abqm.add_link(node_a=site_to_ns_node_id[source_site], rel=ABCPropertyGraph.REL_CONNECTS,
                          node_b=source_cp_id)
            abqm.add_link(node_a=source_cp_id, rel=ABCPropertyGraph.REL_CONNECTS,
                          node_b=link)
            abqm.add_link(node_a=link, rel=ABCPropertyGraph.REL_CONNECTS,
                          node_b=sink_cp_id)
            abqm.add_link(node_a=sink_cp_id, rel=ABCPropertyGraph.REL_CONNECTS,
                          node_b=site_to_ns_node_id[sink_site])

        # link facilities to their sites
        for s, lf in facilities_by_site.items():
            # multiple facilities per site possible
            for fac_sliver in lf:
                ns_list = cbm.get_first_neighbor(node_id=fac_sliver.node_id,
                                                 rel=ABCPropertyGraph.REL_HAS,
                                                 node_label=ABCPropertyGraph.CLASS_NetworkService)

                if not ns_list or not len(ns_list):
                    if self.logger:
                        self.logger.warning(f'Unable to trace facility NetworkService for '
                                            f'facility {fac_sliver.resource_name}, continuing')
                    else:
                        print(f'Unable to trace facility NetworkService for '
                              f'facility {fac_sliver.resource_name}, continuing')
                    continue

                _, fac_props = cbm.get_node_properties(node_id=fac_sliver.node_id)
                new_fac_props = {ABCPropertyGraph.PROP_NAME: fac_props[ABCPropertyGraph.PROP_NAME],
                                 ABCPropertyGraph.PROP_TYPE: fac_props[ABCPropertyGraph.PROP_TYPE],
                                 ABCPropertyGraph.PROP_SITE: s
                                 }
                abqm.add_node(node_id=fac_sliver.node_id, label=ABCPropertyGraph.CLASS_NetworkNode,
                              props=new_fac_props)

                fac_ns_node_id = ns_list[0]
                _, fac_ns_props = cbm.get_node_properties(node_id=fac_ns_node_id)

                # filter down only the needed properties then recreate the structure of facility in ABQM
                new_ns_props = {ABCPropertyGraph.PROP_NAME: fac_ns_props[ABCPropertyGraph.PROP_NAME],
                                ABCPropertyGraph.PROP_TYPE: fac_ns_props[ABCPropertyGraph.PROP_TYPE]
                                }

                abqm.add_node(node_id=fac_ns_node_id, label=ABCPropertyGraph.CLASS_NetworkService,
                              props=new_ns_props)

                abqm.add_link(node_a=fac_sliver.node_id, rel=ABCPropertyGraph.REL_HAS, node_b=fac_ns_node_id)

                fac_ns_cp_list = cbm.get_all_ns_or_link_connection_points(link_id=ns_list[0])
                if not fac_ns_cp_list:
                    if self.logger:
                        self.logger.warning(f'Unable to trace facility ConnectionPoint for '
                                            f'facility {fac_sliver.resource_name}, continuing')
                    else:
                        print(f'Unable to trace facility ConnectionPoint for '
                              f'facility {fac_sliver.resource_name}, continuing')
                    continue

                for fac_cp_node_id in fac_ns_cp_list:
                    _, fac_cp_props = cbm.get_node_properties(node_id=fac_cp_node_id)

                    new_cp_props = {ABCPropertyGraph.PROP_NAME: fac_cp_props[ABCPropertyGraph.PROP_NAME],
                                    ABCPropertyGraph.PROP_TYPE: fac_cp_props[ABCPropertyGraph.PROP_TYPE],
                                    ABCPropertyGraph.PROP_LABELS: fac_cp_props.get(ABCPropertyGraph.PROP_LABELS),
                                    ABCPropertyGraph.PROP_CAPACITIES: fac_cp_props.get(ABCPropertyGraph.PROP_CAPACITIES)
                                    }

                    if not self.DEBUG_FLAG and kwargs['query_level'] != 0:
                        allocated_vlans = self.occupied_vlans(db=db, node_id=fac_sliver.resource_name,
                                                              component_name=fac_cp_node_id, start=start, end=end)

                        if allocated_vlans and len(allocated_vlans):
                            labels = Labels(vlan=allocated_vlans)
                            new_cp_props[ABCPropertyGraph.PROP_LABEL_ALLOCATIONS] = labels.to_json()

                    new_cp_props = {k: v for (k, v) in new_cp_props.items() if v}

                    abqm.add_node(node_id=fac_cp_node_id, label=ABCPropertyGraph.CLASS_ConnectionPoint,
                                  props=new_cp_props)
                    abqm.add_link(node_a=fac_ns_node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=fac_cp_node_id)

                    # trace the link to a switch port/ConnectionPoint and replicate them for simplicity
                    fac_cp_nbs = cbm.get_first_and_second_neighbor(node_id=fac_cp_node_id,
                                                                   rel1=ABCPropertyGraph.REL_CONNECTS,
                                                                   node1_label=ABCPropertyGraph.CLASS_Link,
                                                                   rel2=ABCPropertyGraph.REL_CONNECTS,
                                                                   node2_label=ABCPropertyGraph.CLASS_ConnectionPoint)
                    if len(fac_cp_nbs) == 0 or len(fac_cp_nbs) > 1:
                        if self.logger:
                            self.logger.warning(f'Unable to trace switch port from Facility port '
                                                f'for facility {fac_sliver.resource_name} {fac_cp_nbs}')
                        else:
                            print(f'Unable to trace switch port from Facility port '
                                  f'for facility {fac_sliver.resource_name} {fac_cp_nbs}')
                        continue

                    fac_link_id = fac_cp_nbs[0][0]
                    fac_sp_id = fac_cp_nbs[0][1]

                    _, fac_link_props = cbm.get_node_properties(node_id=fac_link_id)
                    # selectively replicate link properties
                    new_link_props = {ABCPropertyGraph.PROP_NAME: fac_link_props[ABCPropertyGraph.PROP_NAME],
                                      ABCPropertyGraph.PROP_TYPE: fac_link_props[ABCPropertyGraph.PROP_TYPE],
                                      ABCPropertyGraph.PROP_LAYER: fac_link_props[ABCPropertyGraph.PROP_LAYER]
                                      }
                    if kwargs['query_level'] == 0:
                        if fac_link_props.get(ABCPropertyGraph.PROP_CAPACITIES):
                            new_link_props[ABCPropertyGraph.PROP_CAPACITIES] = fac_link_props[
                                ABCPropertyGraph.PROP_CAPACITIES]
                        if fac_link_props.get(ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS):
                            new_link_props[ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS] = fac_link_props[
                                ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS]
                    else:
                        if fac_link_props.get(ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS):
                            new_link_props[ABCPropertyGraph.PROP_CAPACITIES] = fac_link_props[
                                ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS]
                        elif fac_link_props.get(ABCPropertyGraph.PROP_CAPACITIES):
                            new_link_props[ABCPropertyGraph.PROP_CAPACITIES] = fac_link_props[
                                ABCPropertyGraph.PROP_CAPACITIES]
                        if not self.DEBUG_FLAG:
                            occupied_link_capacity = self.occupied_link_capacity(db=db, node_id=fac_link_id,
                                                                                 start=start, end=end)
                            if occupied_link_capacity:
                                new_link_props[ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS] = occupied_link_capacity

                    abqm.add_node(node_id=fac_link_id, label=ABCPropertyGraph.CLASS_Link,
                                  props=new_link_props)
                    try:
                        new_sp_props = abqm.get_node_properties(node_id=fac_sp_id)
                    except PropertyGraphQueryException:
                        # if the node doesn't exist we need to create it (it could have been created in the first pass)
                        _, fac_sp_props = cbm.get_node_properties(node_id=fac_sp_id)
                        new_sp_props = {ABCPropertyGraph.PROP_NAME: fac_sp_props[ABCPropertyGraph.PROP_NAME],
                                        ABCPropertyGraph.PROP_TYPE: fac_sp_props[ABCPropertyGraph.PROP_TYPE],
                                        ABCPropertyGraph.PROP_CAPACITIES: fac_sp_props.get(
                                            ABCPropertyGraph.PROP_CAPACITIES),
                                        ABCPropertyGraph.PROP_LABELS: fac_sp_props.get(ABCPropertyGraph.PROP_LABELS)
                                        }
                        new_sp_props = {k: v for (k, v) in new_sp_props.items() if v}
                        abqm.add_node(node_id=fac_sp_id, label=ABCPropertyGraph.CLASS_ConnectionPoint,
                                      props=new_sp_props)

                    # link these together
                    abqm.add_link(node_a=fac_cp_node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=fac_link_id)
                    abqm.add_link(node_a=fac_link_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=fac_sp_id)
                    abqm.add_link(node_a=fac_sp_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=site_to_ns_node_id[s])

        return abqm



from typing import Tuple
from fabric_cf.actor.core.common.constants import Constants


class TokenException(Exception):
    """
    Token exception
    """


class FabricToken:
    """
    Represents the Fabric Token issues by Credential Manager
    """
    def __init__(self, *, decoded_token: dict, token_hash: str):
        self.decoded_token = decoded_token
        self.hash = token_hash

    @property
    def token_hash(self):
        return self.hash

    @property
    def token(self):
        return self.decoded_token

    @property
    def uuid(self) -> str:
        return self.decoded_token.get(Constants.UUID)

    @property
    def subject(self) -> str:
        """
        Get subject
        @return subject
        """
        return self.decoded_token.get(Constants.CLAIMS_SUB)

    @property
    def email(self) -> str:
        """
        Get email
        @return email
        """
        return self.decoded_token.get(Constants.CLAIMS_EMAIL)

    @property
    def projects(self) -> list or None:
        """
        Get projects
        @return projects
        """
        return self.decoded_token.get(Constants.CLAIMS_PROJECTS)

    def __str__(self):
        return f"Token: {self.decoded_token}"

    @property
    def first_project(self) -> Tuple[str or None, str or None, str or None]:
        if self.projects is None or len(self.projects) == 0:
            return None, None, None

        return self.projects[0].get(Constants.UUID), self.projects[0].get(Constants.TAGS),\
               self.projects[0].get(Constants.NAME)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import List

from fim.slivers.base_sliver import BaseSliver
from fim.slivers.network_service import NetworkServiceSliver
from fim.slivers.topology_diff import WhatsModifiedFlag

from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.exceptions import AuthorityException
from fabric_cf.actor.core.core.unit import Unit
from fabric_cf.actor.core.core.unit_set import UnitSet
from fabric_cf.actor.core.kernel.resource_set import ResourceSet
from fabric_cf.actor.core.policy.resource_control import ResourceControl
from fabric_cf.actor.core.util.resource_type import ResourceType


class NetworkServiceControl(ResourceControl):
    """
    Resource Control for Network Service
    """
    @staticmethod
    def __has_sliver_changed(current: NetworkServiceSliver, requested: NetworkServiceSliver):
        diff = current.diff(other_sliver=requested)
        if diff is None:
            return False

        if (diff.added.interfaces is not None and len(diff.added.interfaces)) or \
                (diff.removed.interfaces is not None and len(diff.removed.interfaces)) or \
                (diff.modified.interfaces is not None and len(diff.modified.interfaces)):
            return True

        if diff.modified is not None and diff.modified.services is not None:
            for new_ns, flag in diff.modified.services:
                if flag & WhatsModifiedFlag.LABELS or flag & WhatsModifiedFlag.CAPACITIES:
                    return True
        return False

    def assign(self, *, reservation: ABCAuthorityReservation, delegation_name: str,
               graph_node: BaseSliver, existing_reservations: List[ABCReservationMixin]) -> ResourceSet:
        """
        Assign a reservation
        :param reservation: reservation
        :param delegation_name: Name of delegation serving the request
        :param graph_node: ARM Graph Node serving the reservation
        :param existing_reservations: Existing Reservations served by the same ARM node
        :return: ResourceSet with updated sliver annotated with properties
        :raises: AuthorityException in case the request cannot be satisfied
        """
        reservation.set_send_with_deficit(value=True)

        requested = reservation.get_requested_resources().get_sliver()
        self.__dump_sliver(sliver=requested)
        if not isinstance(requested, NetworkServiceSliver):
            raise AuthorityException(f"Invalid resource type {requested.get_type()}")

        current = reservation.get_resources()

        resource_type = ResourceType(resource_type=str(requested.get_type()))

        gained = None
        modified = None
        lost = None
        if current is None:
            self.logger.debug("check if sliver can be provisioned")
            # FIXME Add validation to check ticketed sliver against ARM
            # Refer Network Node Control for reference

            self.logger.debug(f"Slice properties: {reservation.get_slice().get_config_properties()}")
            unit = Unit(rid=reservation.get_reservation_id(), slice_id=reservation.get_slice_id(),
                        actor_id=self.authority.get_guid(), sliver=requested, rtype=resource_type,
                        properties=reservation.get_slice().get_config_properties())
            gained = UnitSet(plugin=self.authority.get_plugin(), units={unit.reservation_id: unit})
        else:
            current_sliver = current.get_sliver()
            if self.__has_sliver_changed(current=current_sliver, requested=requested):
                unit = Unit(rid=reservation.get_reservation_id(), slice_id=reservation.get_slice_id(),
                            actor_id=self.authority.get_guid(), sliver=requested, rtype=resource_type,
                            properties=reservation.get_slice().get_config_properties())
                modified = UnitSet(plugin=self.authority.get_plugin(), units={unit.reservation_id: unit})
            else:
                return current

        result = ResourceSet(gained=gained, modified=modified, lost=lost, rtype=resource_type)
        result.set_sliver(sliver=requested)
        return result

    def revisit(self, *, reservation: ABCReservationMixin):
        return

    def free(self, *, uset: dict):
        if uset is not None:
            for u in uset.values():
                self.logger.debug(f"Freeing 1 unit {u}")

    def __dump_sliver(self, *, sliver: NetworkServiceSliver):
        self.logger.debug(f"Network Service Sliver# {sliver}")
        if sliver.interface_info is not None:
            for ifs in sliver.interface_info.interfaces.values():
                self.logger.debug(f"Ifs# {ifs}")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import concurrent.futures
import logging
import multiprocessing
import os
import threading
import time
import traceback

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import AuthorityException
from fabric_cf.actor.core.plugins.handlers.config_token import ConfigToken
from fabric_cf.actor.core.plugins.handlers.handler_processor import HandlerProcessor
from fabric_cf.actor.core.util.log_helper import LogHelper
from fabric_cf.actor.core.util.reflection_utils import ReflectionUtils

process_pool_logger = None


class AnsibleHandlerProcessor(HandlerProcessor):
    """
    Ansible Handler Processor
    """
    MAX_WORKERS = 10

    def __init__(self):
        super().__init__()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.log_config = GlobalsSingleton.get().get_log_config()
        self.executor = None
        self.process_pool_manager = None
        self.process_pool_lock = None
        self.__setup_process_pool()
        self.futures = []
        self.thread = None
        self.future_lock = threading.Condition()
        self.stopped = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['plugin']
        del state['initialized']
        del state['lock']
        del state['executor']
        del state['process_pool_manager']
        del state['process_pool_lock']
        del state['futures']
        del state['thread']
        del state['future_lock']
        del state['stopped']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.plugin = None
        self.initialized = False
        self.lock = threading.Lock()
        self.__setup_process_pool()
        self.futures = []
        self.thread = None
        self.future_lock = threading.Condition()
        self.stopped = False

    def __setup_process_pool(self):
        log_dir = self.log_config.get(Constants.PROPERTY_CONF_LOG_DIRECTORY, ".")
        log_file = self.log_config.get(Constants.PROPERTY_CONF_HANDLER_LOG_FILE, "handler.log")
        log_level = self.log_config.get(Constants.PROPERTY_CONF_LOG_LEVEL, logging.DEBUG)
        log_retain = int(self.log_config.get(Constants.PROPERTY_CONF_LOG_RETAIN, 50))
        log_size = int(self.log_config.get(Constants.PROPERTY_CONF_LOG_SIZE, 5000000))
        logger = self.log_config.get(Constants.PROPERTY_CONF_LOGGER, "handler")
        logger = f"{logger}-handler"
        self.executor = concurrent.futures.ProcessPoolExecutor(max_workers=self.MAX_WORKERS,
                                                               initializer=AnsibleHandlerProcessor.process_pool_initializer,
                                                               initargs=(log_dir, log_file, log_level, log_retain,
                                                                         log_size, logger))
        self.process_pool_manager = multiprocessing.Manager()
        self.process_pool_lock = self.process_pool_manager.Lock()

    def start(self):
        """
        Start the Future Processor Thread
        """
        self.thread = threading.Thread(target=self.process_futures)
        self.thread.setName("FutureProcessor")
        self.thread.setDaemon(True)
        self.thread.start()

    def shutdown(self):
        """
        Shutdown Process Pool, Future Processor
        """
        try:
            self.lock.acquire()
            self.executor.shutdown(wait=True)
            self.stopped = True
            for f, u in self.futures:
                f.cancel()

            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the future thread is running. Interrupting it")
                try:
                    with self.future_lock:
                        self.future_lock.notify_all()
                    temp.join()
                except Exception as e:
                    self.logger.error("Could not join future thread {}".format(e))
                    self.logger.error(traceback.format_exc())
                finally:
                    self.lock.release()
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
        finally:
            if self.lock.locked():
                self.lock.release()

    def set_logger(self, *, logger):
        self.logger = logger

    def queue_future(self, future: concurrent.futures.Future, unit: ConfigToken):
        """
        Queue an future on Future Processor timer queue
        @param future
        @param unit Unit being processed; This is needed
        """
        with self.future_lock:
            self.futures.append((future, unit))
            self.logger.debug("Added future to Future queue")
            self.future_lock.notify_all()

    def invoke_handler(self, unit: ConfigToken, operation: str, data: dict = None):
        try:
            # clean restart
            if unit is None:
                if len(self.config_mappings) == 0:
                    return
                handler = list(self.config_mappings.values())[0]
            else:
                handler = self.config_mappings.get(str(unit.get_resource_type()), None)
            if handler is None:
                raise AuthorityException(f"No handler found for resource type {unit.get_resource_type()}")

            future = self.executor.submit(self.process_pool_main, operation, handler.get_class_name(),
                                          handler.get_module_name(), handler.get_properties(), unit, data,
                                          self.process_pool_lock)

            self.queue_future(future=future, unit=unit)
            if unit:
                self.logger.debug(f"Handler operation {operation} scheduled for Resource Type: {unit.get_resource_type()} "
                                  f"Unit: {unit.get_id()} Reservation: {unit.get_reservation_id()}")

        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
            result = {Constants.PROPERTY_TARGET_NAME: operation,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_EXCEPTION,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}

            self.handler_complete(unit=unit, properties=result, old_unit=unit)
        finally:
            self.logger.info(f"Executing {operation} completed")

    def create(self, unit: ConfigToken):
        self.invoke_handler(unit=unit, operation=Constants.TARGET_CREATE)

    def modify(self, unit: ConfigToken):
        self.invoke_handler(unit=unit, operation=Constants.TARGET_MODIFY)

    def poa(self, unit: ConfigToken, data: dict):
        self.invoke_handler(unit=unit, operation=Constants.TARGET_POA, data=data)

    def delete(self, unit: ConfigToken):
        self.invoke_handler(unit=unit, operation=Constants.TARGET_DELETE)

    def clean_restart(self):
        self.invoke_handler(unit=None, operation=Constants.TARGET_CLEAN_RESTART)

    def handler_complete(self, properties: dict, unit: ConfigToken, old_unit: ConfigToken):
        try:
            self.lock.acquire()
            self.logger.debug(f"Properties: {properties} Unit: {unit}")
            target = properties.get(Constants.PROPERTY_TARGET_NAME)
            if target != Constants.TARGET_POA:
                # Copy the sliver from the Unit to
                old_unit.update_sliver(sliver=unit.get_sliver())
            self.plugin.configuration_complete(token=old_unit, properties=properties)
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())
        finally:
            self.lock.release()

    def process(self, future: concurrent.futures.Future, old_unit: ConfigToken):
        if old_unit is None:
            return
        try:
            self.logger.debug(f"Handler Execution completed Result: {future.result()}")
            if future.exception() is not None:
                self.logger.error(f"Exception occurred while executing the handler: {future.exception()}")
            properties, unit = future.result()
            self.handler_complete(properties=properties, unit=unit, old_unit=old_unit)
        except Exception as e:
            self.logger.error(f"Exception occurred {e}")
            self.logger.error(traceback.format_exc())

    def process_futures(self):
        while True:
            done = []

            with self.future_lock:

                while len(self.futures) == 0 and not self.stopped:
                    try:
                        self.future_lock.wait()
                    except InterruptedError as e:
                        self.logger.info("Future Processor thread interrupted. Exiting")
                        return

                if self.stopped:
                    self.logger.info("Future Processor exiting")
                    return

                if len(self.futures) > 0:
                    try:
                        for f, u in self.futures:
                            if f.done():
                                done.append((f, u))

                        for x in done:
                            self.futures.remove(x)
                    except Exception as e:
                        self.logger.error(f"Error while adding future to future queue! e: {e}")
                        self.logger.error(traceback.format_exc())

                self.future_lock.notify_all()

            if len(done) > 0:
                self.logger.debug(f"Processing {len(done)} futures")
                for f, u in done:
                    try:
                        self.process(future=f, old_unit=u)
                    except Exception as e:
                        self.logger.error(f"Error while processing event {type(f)}, {e}")
                        self.logger.error(traceback.format_exc())

                done.clear()

            time.sleep(5)

    @staticmethod
    def process_pool_main(operation: str, handler_class: str, handler_module: str, properties: dict,
                          unit: ConfigToken, data: dict, process_lock: multiprocessing.Lock):
        global process_pool_logger
        handler_class = ReflectionUtils.create_instance_with_params(module_name=handler_module,
                                                                    class_name=handler_class)
        handler_obj = handler_class(process_pool_logger, properties, process_lock)

        if operation == Constants.TARGET_CREATE:
            return handler_obj.create(unit)
        elif operation == Constants.TARGET_DELETE:
            return handler_obj.delete(unit)
        elif operation == Constants.TARGET_MODIFY:
            return handler_obj.modify(unit)
        elif operation == Constants.TARGET_POA:
            return handler_obj.poa(unit, data)
        elif operation == Constants.TARGET_CLEAN_RESTART:
            return handler_obj.clean_restart()
        else:
            process_pool_logger.error("Invalid operation")
            result = {Constants.PROPERTY_TARGET_NAME: operation,
                      Constants.PROPERTY_TARGET_RESULT_CODE: Constants.RESULT_CODE_EXCEPTION,
                      Constants.PROPERTY_ACTION_SEQUENCE_NUMBER: 0}
            return result, unit

    @staticmethod
    def process_pool_initializer(log_dir: str, log_file: str, log_level, log_retain: int, log_size: int,
                                 logger: str):
        global process_pool_logger
        if process_pool_logger is None:
            log_file = f"{os.getpid()}-{log_file}"
            process_pool_logger = LogHelper.make_logger(log_dir=log_dir, log_file=log_file, log_level=log_level,
                                                        log_retain=log_retain, log_size=log_size, logger=logger)
